@@ -1,5 +1,5 @@
 /*
- * Copyright 1999, 2000, 2001, 2002, 2007 by Paul Mattes.
+ * Copyright 1999-2008 by Paul Mattes.
  *  Permission to use, copy, modify, and distribute this software and its
  *  documentation for any purpose and without fee is hereby granted,
  *  provided that the above copyright notice appear in all copies and that
@@ -27,7 +27,9 @@
 #include <errno.h>
 #include <X11/keysym.h>
 
+#if !defined(_MSC_VER) /*[*/
 #include <sys/time.h>
+#endif /*]*/
 #include <sys/types.h>
 #if defined(_WIN32) /*[*/
 #include <windows.h>
@@ -320,12 +322,14 @@ static struct {
 	 * The following are, umm, hacks to allow symbolic names for
 	 * control codes.
 	 */
+#if !defined(_WIN32) /*[*/
 	{ "BackSpace", 0x08 },
 	{ "Tab", 0x09 },
 	{ "Linefeed", 0x0a },
 	{ "Return", 0x0d },
 	{ "Escape", 0x1b },
 	{ "Delete", 0x7f },
+#endif /*]*/
 
 	{ (char *)NULL, NoSymbol }
 };
@@ -368,8 +372,8 @@ ms_ts(unsigned long long *u)
 	GetSystemTimeAsFileTime(&t);
 	memcpy(u, &t, sizeof(unsigned long long));
 
-	/* Divide by 10 to get ms. */
-	*u /= 10ULL;
+	/* Divide by 10,000 to get ms. */
+	*u /= 10000ULL;
 }
 #endif /*]*/
 
@@ -459,52 +463,15 @@ RemoveTimeOut(unsigned long timer)
 	}
 }
 
-/* Basic doubly-linked lists. */
-typedef struct _dkl {
-    	struct _dkl *next;
-	struct _dkl *prev;
-	void *payload;
-} dkl_t;
-
-void
-dkl_init(dkl_t *elt, void *payload)
-{
-    	elt->next = elt->prev = elt;
-	elt->payload = payload;
-}
-
-void
-dkl_append(dkl_t *head, dkl_t *elt)
-{
-    	elt->next = head;
-	elt->prev = head->prev;
-	head->prev->next = elt;
-	head->prev = elt;
-}
-
-void
-dkl_remove(dkl_t *elt)
-{
-	elt->prev->next = elt->next;
-	elt->next->prev = elt->prev;
-}
-
-int
-dkl_end(dkl_t *head, dkl_t *elt)
-{
-    	return (elt->next == head);
-}
-
 /* Input events. */ 
 typedef struct input {  
-    	dkl_t dkl;
+        struct input *next;
         int source; 
         int condition;
         void (*proc)(void);
 } input_t;          
-static dkl_t inputs = { &inputs, &inputs, NULL };
+static input_t *inputs = (input_t *)NULL;
 static Boolean inputs_changed = False;
-int num_inputs = 0;
 
 unsigned long
 AddInput(int source, void (*fn)(void))
@@ -515,10 +482,9 @@ AddInput(int source, void (*fn)(void))
 	ip->source = source;
 	ip->condition = InputReadMask;
 	ip->proc = fn;
-	dkl_init(&ip->dkl, ip);
-	dkl_append(&inputs, &ip->dkl);
+	ip->next = inputs;
+	inputs = ip;
 	inputs_changed = True;
-	num_inputs++;
 	return (unsigned long)ip;
 }
 
@@ -534,10 +500,9 @@ AddExcept(int source, void (*fn)(void))
 	ip->source = source;
 	ip->condition = InputExceptMask;
 	ip->proc = fn;
-	dkl_init(&ip->dkl, ip);
-	dkl_append(&inputs, &ip->dkl);
+	ip->next = inputs;
+	inputs = ip;
 	inputs_changed = True;
-	num_inputs++;
 	return (unsigned long)ip;
 #endif /*]*/
 }
@@ -552,61 +517,34 @@ AddOutput(int source, void (*fn)(void))
 	ip->source = source;
 	ip->condition = InputWriteMask;
 	ip->proc = fn;
-	dkl_init(&ip->dkl, ip);
-	dkl_append(&inputs, &ip->dkl);
+	ip->next = inputs;
+	inputs = ip;
 	inputs_changed = True;
-	num_inputs++;
 	return (unsigned long)ip;
 }
 #endif /*]*/
 
-input_t *
-FirstInput(void)
-{
-    	return (input_t *)inputs.next->payload;
-}
-
-input_t *
-NextInput(input_t *ip)
-{
-    	return (input_t *)((ip->dkl.next)->payload);
-}
-
-int
-LastInput(input_t *ip)
-{
-    	return (ip == NULL);
-}
-
 void
 RemoveInput(unsigned long id)
 {
 	input_t *ip;
+	input_t *prev = (input_t *)NULL;
 
-	for (ip = FirstInput(); !LastInput(ip); ip = NextInput(ip)) {
+	for (ip = inputs; ip != (input_t *)NULL; ip = ip->next) {
 		if (ip == (input_t *)id)
 			break;
+		prev = ip;
 	}
 	if (ip == (input_t *)NULL)
 		return;
-	dkl_remove(&ip->dkl);
+	if (prev != (input_t *)NULL)
+		prev->next = ip->next;
+	else
+		inputs = ip->next;
 	Free(ip);
 	inputs_changed = True;
-	num_inputs--;
 }
 
-#if defined(_WIN32) /*[*/
-/* Move an input to the tail of the list. */
-void
-TailInput(input_t *ip)
-{
-    	if (!LastInput(ip)) {
-	    	dkl_remove(&ip->dkl);
-		dkl_append(&inputs, &ip->dkl);
-	}
-}
-#endif /*]*/
-
 #if !defined(_WIN32) /*[*/
 /*
  * Modify the passed-in parameters so they reflect the values needed for
@@ -619,7 +557,7 @@ select_setup(int *nfds, fd_set *readfds, fd_set *writefds,
 	input_t *ip;
 	int r = 0;
 
-	for (ip = FirstInput(); !LastInput(ip); ip = NextInput(ip)) {
+	for (ip = inputs; ip != (input_t *)NULL; ip = ip->next) {
 		if ((unsigned long)ip->condition & InputReadMask) {
 			FD_SET(ip->source, readfds);
 			if (ip->source >= *nfds)
@@ -707,7 +645,7 @@ process_events(Boolean block)
 	FD_ZERO(&wfds);
 	FD_ZERO(&xfds);
 #endif /*]*/
-	for (ip = FirstInput(); !LastInput(ip); ip = NextInput(ip)) {
+	for (ip = inputs; ip != (input_t *)NULL; ip = ip->next) {
 		if ((unsigned long)ip->condition & InputReadMask) {
 #if defined(_WIN32) /*[*/
 			ha[nha++] = (HANDLE)ip->source;
@@ -779,11 +717,11 @@ process_events(Boolean block)
 	}
 	inputs_changed = False;
 #if defined(_WIN32) /*[*/
-	for (i = 0, ip = FirstInput(); !LastInput(ip); ip = ip_next, i++) {
+	for (i = 0, ip = inputs; ip != (input_t *)NULL; ip = ip_next, i++) {
 #else /*][*/
-	for (ip = FirstInput(); !LastInput(ip); ip = ip_next) {
+	for (ip = inputs; ip != (input_t *)NULL; ip = ip_next) {
 #endif /*]*/
-		ip_next = NextInput(ip);
+		ip_next = ip->next;
 		if (((unsigned long)ip->condition & InputReadMask) &&
 #if defined(_WIN32) /*[*/
 		    ret == WAIT_OBJECT_0 + i) {
@@ -794,9 +732,6 @@ process_events(Boolean block)
 			processed_any = True;
 			if (inputs_changed)
 				goto retry;
-#if defined(_WIN32) /*[*/
-			TailInput(ip);
-#endif /*]*/
 		}
 #if !defined(_WIN32) /*[*/
 		if (((unsigned long)ip->condition & InputWriteMask) &&
