@@ -1,6 +1,5 @@
 /*
- * Modifications and original code Copyright 1993, 1994, 1995, 1996,
- *    2000, 2001, 2002, 2003, 2004, 2005, 2006, 2007 by Paul Mattes.
+ * Modifications and original code Copyright 1993-2009 by Paul Mattes.
  * Original X11 Port Copyright 1990 by Jeff Sparkes.
  *   Permission to use, copy, modify, and distribute this software and its
  *   documentation for any purpose and without fee is hereby granted,
@@ -52,6 +51,7 @@
 #include "telnetc.h"
 #include "togglesc.h"
 #include "trace_dsc.h"
+#include "utf8c.h"
 #include "utilc.h"
 #include "xioc.h"
 
@@ -62,6 +62,12 @@
 #endif /*]*/
 #endif /*]*/
 
+#if defined(_WIN32) /*[*/
+#include <windows.h>
+#include "winversc.h"
+#include "windirsc.h"
+#endif /*]*/
+
 static void interact(void);
 static void stop_pager(void);
 
@@ -71,7 +77,9 @@ static char *completion_entry(const char *, int);
 #endif /*]*/
 
 /* Pager state. */
+#if !defined(_WIN32) /*[*/
 static FILE *pager = NULL;
+#endif /*]*/
 
 #if !defined(_WIN32) /*[*/
 /* Base keymap. */
@@ -144,103 +152,56 @@ static char *base_3270_keymap =
 #else /*][*/
 
 /* Base keymap. */
-static char *base_keymap1 =
-"Alt Ctrl<Key>]: Key(0x1d)\n"
-"Ctrl<Key>]: Escape\n"
-"Alt <Key>c: Clear\n"
-"Alt <Key>i: Insert\n"
-"Alt <Key>r: Reset\n"
-"Alt <Key>l: Redraw\n"
-"Alt <Key>m: Compose\n"
-"Alt <Key>^: Key(notsign)\n"
-"<Key>Delete: Delete\n"
-"<Key>Up: Up\n"
-"<Key>Down: Down\n"
-"<Key>Left: Left\n"
-"<Key>Right: Right\n"
-"<Key>Home: Home\n"
-"Alt <Key>1: PA(1)\n"
-"Alt <Key>2: PA(2)\n"
-"Alt <Key>3: PA(3)\n"
-"<Key>Insert: ToggleInsert\n"
-"<Key>Numpad0: Key(0)\n"
-"<Key>Numpad1: Key(1)\n"
-"<Key>Numpad2: Key(2)\n"
-"<Key>Numpad3: Key(3)\n"
-"<Key>Numpad4: Key(4)\n"
-"<Key>Numpad5: Key(5)\n"
-"<Key>Numpad6: Key(6)\n"
-"<Key>Numpad7: Key(7)\n"
-"<Key>Numpad8: Key(8)\n"
-"<Key>Numpad9: Key(9)\n"
-"<Key>Multiply: Key(*)\n"
-"<Key>Add: Key(+)\n"
-"<Key>Separator: Key(|)\n"
-"<Key>Subtract: Key(-)\n"
-"<Key>Decimal: Key(.)\n"
-"<Key>Divide: Key(/)\n";
-static char *base_keymap2 =
-"Shift <Key>F1: PF(13)\n"
-"<Key>F1: PF(1)\n"
-"Shift <Key>F2: PF(14)\n"
-"<Key>F2: PF(2)\n"
-"Shift <Key>F3: PF(15)\n"
-"<Key>F3: PF(3)\n"
-"Shift <Key>F4: PF(16)\n"
-"<Key>F4: PF(4)\n"
-"Shift <Key>F5: PF(17)\n"
-"<Key>F5: PF(5)\n"
-"Shift <Key>F6: PF(18)\n"
-"<Key>F6: PF(6)\n";
-static char *base_keymap3 =
-"Shift <Key>F7: PF(19)\n"
-"<Key>F7: PF(7)\n"
-"Shift <Key>F8: PF(20)\n"
-"<Key>F8: PF(8)\n"
-"Shift <Key>F9: PF(21)\n"
-"<Key>F9: PF(9)\n"
-"Shift <Key>F10: PF(22)\n"
-"<Key>F10: PF(10)\n"
-"Shift <Key>F11: PF(23)\n"
-"<Key>F11: PF(11)\n"
-"Shift <Key>F12: PF(24)\n"
-"<Key>F12: PF(12)\n"
-"<Key>F13: PF(13)\n"
-"<Key>F14: PF(14)\n"
-"<Key>F15: PF(15)\n"
-"<Key>F16: PF(16)\n"
-"<Key>F17: PF(17)\n"
-"<Key>F18: PF(18)\n"
-"<Key>F19: PF(19)\n"
-"<Key>F20: PF(20)\n"
-"<Key>F21: PF(21)\n"
-"<Key>F22: PF(22)\n"
-"<Key>F23: PF(23)\n"
-"<Key>F24: PF(24)\n";
+static char *base_keymap =
+       "Alt <Key>1:      PA(1)\n"
+       "Alt <Key>2:      PA(2)\n"
+       "Alt <Key>3:      PA(3)\n"
+       "Alt <Key>^:      Key(notsign)\n"
+       "Alt <Key>c:      Clear\n"
+       "Alt <Key>l:      Redraw\n"
+       "Alt <Key>m:      Compose\n"
+       "Alt <Key>p:      PrintText\n"
+      "Ctrl <Key>]:      Escape\n"
+     "Shift <Key>F1:     PF(13)\n"
+     "Shift <Key>F2:     PF(14)\n"
+     "Shift <Key>F3:     PF(15)\n"
+     "Shift <Key>F4:     PF(16)\n"
+     "Shift <Key>F5:     PF(17)\n"
+     "Shift <Key>F6:     PF(18)\n"
+     "Shift <Key>F7:     PF(19)\n"
+     "Shift <Key>F8:     PF(20)\n"
+     "Shift <Key>F9:     PF(21)\n"
+     "Shift <Key>F10:    PF(22)\n"
+     "Shift <Key>F11:    PF(23)\n"
+     "Shift <Key>F12:    PF(24)\n"
+     "Shift <Key>ESCAPE: Key(0x1d)\n";
 
 /* Base keymap, 3270 mode. */
 static char *base_3270_keymap =
-"Ctrl <Key>[: Escape\n"
-"Ctrl <Key>c: Clear\n"
-"Ctrl <Key>r: Reset\n"
-"Ctrl <Key>l: Redraw\n"
-"Shift Ctrl <Key>i: BackTab\n"
-"Ctrl <Key>i: Tab\n"
-"<Key> Delete: Delete\n"
-"Ctrl <Key>h: Erase\n"
-"Ctrl <Key>j: Newline\n"
-"Ctrl <Key>m: Enter\n"
-"<Key> Home: Home\n"
-"Ctrl <Key>a: Attn\n"
-"Alt <Key>a: Attn\n"
-"Shift <Key> End: EraseEOF\n"
-"<Key> End: FieldEnd\n"
-"Ctrl <Key>v: Paste\n"
-"Ctrl <Key>f: FieldMark\n"
-"Ctrl <Key>d: Dup\n"
-"Ctrl <Key>u: DeleteField\n"
-"Shift <Key>Left: PreviousWord\n"
-"Shift <Key>Right: NextWord\n";
+      "Ctrl <Key>a:      Attn\n"
+       "Alt <Key>a:      Attn\n"
+      "Ctrl <Key>c:      Clear\n"
+      "Ctrl <Key>d:      Dup\n"
+       "Alt <Key>d:      Dup\n"
+      "Ctrl <Key>f:      FieldMark\n"
+       "Alt <Key>f:      FieldMark\n"
+      "Ctrl <Key>h:      Erase\n"
+       "Alt <Key>i:      Insert\n"
+      "Ctrl <Key>i:      Tab\n"
+      "Ctrl <Key>j:      Newline\n"
+      "Ctrl <Key>l:      Redraw\n"
+      "Ctrl <Key>m:      Enter\n"
+      "Ctrl <Key>r:      Reset\n"
+       "Alt <Key>r:      Reset\n"
+      "Ctrl <Key>u:      DeleteField\n"
+      "Ctrl <Key>v:      Paste\n"
+           "<Key>INSERT: ToggleInsert\n"
+     "Shift <Key>TAB:    BackTab\n"
+           "<Key>BACK:   Erase\n"
+     "Shift <Key>END:    EraseEOF\n"
+           "<Key>END:    FieldEnd\n"
+     "Shift <Key>LEFT:   PreviousWord\n"
+     "Shift <Key>RIGHT:  NextWord\n";
 #endif /*]*/
 
 Boolean any_error_output = False;
@@ -248,6 +209,11 @@ Boolean escape_pending = False;
 Boolean stop_pending = False;
 Boolean dont_return = False;
 
+#if defined(_WIN32) /*[*/
+char *instdir = NULL;
+char myappdata[MAX_PATH];
+#endif /*]*/
+
 void
 usage(char *msg)
 {
@@ -257,6 +223,35 @@ usage(char *msg)
 		programname);
 }
 
+#if defined(_WIN32) /*[*/
+/*
+ * Figure out the install directory and our data directory.
+ */
+static void
+save_dirs(char *argv0)
+{
+    	char *bsl;
+
+	/* Extract the installation directory from argv[0]. */
+	bsl = strrchr(argv0, '\\');
+	if (bsl != NULL) {
+	    	instdir = NewString(argv0);
+		instdir[(bsl - argv0) + 1] = '\0';
+	} else
+	    	instdir = "";
+
+	/* Figure out the application data directory. */
+	if (get_dirs(NULL, myappdata, "wc3270") < 0)
+	    	x3270_exit(1);
+
+	/*
+	 * Create the application data directory, in case we are being run
+	 * by someone other than the user that installed us.
+	 */
+	(void) _mkdir(myappdata);
+}
+#endif /*]*/
+
 /* Callback for connection state changes. */
 static void
 main_connect(Boolean ignored)
@@ -311,18 +306,28 @@ main(int argc, char *argv[])
 {
 	const char	*cl_hostname = CN;
 
-	printf("%s\n\n"
-		"Copyright 1989-2007 by Paul Mattes, GTRC and others.\n"
-		"Type 'show copyright' for full copyright information.\n"
-		"Type 'help' for help information.\n\n",
-		build);
+#if defined(_WIN32) /*[*/
+	(void) get_version_info();
+	(void) save_dirs(argv[0]);
+#endif /*]*/
 
 	add_resource("keymap.base",
-	    xs_buffer("%s%s%s", base_keymap1, base_keymap2, base_keymap3));
+#if defined(_WIN32) /*[*/
+	    base_keymap
+#else /*][*/
+	    xs_buffer("%s%s%s", base_keymap1, base_keymap2, base_keymap3)
+#endif /*]*/
+	    );
 	add_resource("keymap.base.3270", NewString(base_3270_keymap));
 
 	argc = parse_command_line(argc, (const char **)argv, &cl_hostname);
 
+	printf("%s\n\n"
+		"Copyright 1989-2009 by Paul Mattes, GTRC and others.\n"
+		"Type 'show copyright' for full copyright information.\n"
+		"Type 'help' for help information.\n\n",
+		build);
+
 	if (charset_init(appres.charset) != CS_OKAY) {
 		xs_warning("Cannot find charset \"%s\"", appres.charset);
 		(void) charset_init(CN);
@@ -445,7 +450,7 @@ main(int argc, char *argv[])
  * c3270 will stop before the next prompt is printed.
  */
 static void
-running_sigtstp_handler(int ignored unused)
+running_sigtstp_handler(int ignored _is_unused)
 {
 	signal(SIGTSTP, SIG_IGN);
 	stop_pending = True;
@@ -458,7 +463,7 @@ running_sigtstp_handler(int ignored unused)
  * of input before resuming the connection.
  */
 static void
-prompt_sigtstp_handler(int ignored unused)
+prompt_sigtstp_handler(int ignored _is_unused)
 {
 	if (CONNECTED)
 		dont_return = True;
@@ -643,11 +648,13 @@ start_pager(void)
 static void
 stop_pager(void)
 {
+#if !defined(_WIN32) /*[*/
 	if (pager != NULL) {
 		if (pager != stdout)
 			pclose(pager);
 		pager = NULL;
 	}
+#endif /*]*/
 }
 
 #if defined(HAVE_LIBREADLINE) /*[*/
@@ -657,7 +664,7 @@ static char **next_match;
 
 /* Generate a match list. */
 static char **
-attempted_completion(char *text, int start, int end)
+attempted_completion(const char *text, int start, int end)
 {
 	char *s;
 	int i, j;
@@ -847,8 +854,44 @@ status_dump(void)
 	action_output("%s %s", get_message("terminalName"), termtype);
 	if (connected_lu != CN && connected_lu[0])
 		action_output("%s %s", get_message("luName"), connected_lu);
-	action_output("%s %s", get_message("characterSet"),
-	    get_charset_name());
+	action_output("%s %s (%s)", get_message("characterSet"),
+	    get_charset_name(),
+#if defined(X3270_DBCS) /*[*/
+	    dbcs? "DBCS": "SBCS"
+#else /*][*/
+	    "SBCS"
+#endif /*]*/
+	    );
+#if defined(X3270_DBCS) /*[*/
+	if (dbcs)
+		action_output("%s %ld+%ld",
+			get_message("hostCodePage"),
+			cgcsgid & 0xffff,
+			cgcsgid_dbcs & 0xffff);
+	else
+#endif /*]*/
+		action_output("%s %ld",
+			get_message("hostCodePage"),
+			cgcsgid & 0xffff);
+#if !defined(_WIN32) /*[*/
+	action_output("%s %s", get_message("localeCodeset"), locale_codeset);
+	action_output("%s DBCS %s, wide curses %s",
+		get_message("buildOpts"),
+#if defined(X3270_DBCS) /*[*/
+		get_message("buildEnabled"),
+#else /*][*/
+		get_message("buildDisabled"),
+#endif /*]*/
+#if defined(CURSES_WIDE) /*[*/
+		get_message("buildEnabled")
+#else /*][*/
+		get_message("buildDisabled")
+#endif /*]*/
+		);
+#else /*][*/
+	action_output("%s OEM %d ANSI %d", get_message("windowsCodePage"),
+		windows_cp, GetACP());
+#endif /*]*/
 	if (appres.key_map) {
 		action_output("%s %s", get_message("keyboardMap"),
 		    appres.key_map);
@@ -963,39 +1006,36 @@ status_dump(void)
 static void
 copyright_dump(void)
 {
-	printf(
-"\n"
-"Modifications and original code Copyright 1993, 1994, 1995, 1996,\n"
-" 2000, 2001, 2002, 2003, 2004, 2005, 2006, 2007 by Paul Mattes.\n"
-"Original X11 Port Copyright 1990 by Jeff Sparkes.\n"
-"DFT File Transfer Code Copyright October 1995 by Dick Altenbern.\n"
-"RPQNAMES Code Copyright 2004, 2005 by Don Russell.\n");
-	printf(
-"  Permission to use, copy, modify, and distribute this software and its\n"
-"  documentation for any purpose and without fee is hereby granted,\n"
-"  provided that the above copyright notice appear in all copies and that\n"
-"  both that copyright notice and this permission notice appear in\n"
-"  supporting documentation.\n");
-	printf(
-"\n"
-"Copyright 1989 by Georgia Tech Research Corporation, Atlanta, GA 30332.\n"
-"  All Rights Reserved.  GTRC hereby grants public use of this software.\n"
-"  Derivative works based on this software must incorporate this copyright\n"
-"  notice.\n"
-"\n"
+	action_output(" ");
+	action_output("Modifications and original code Copyright 1993-2009 by Paul Mattes.");
+	action_output("Original X11 Port Copyright 1990 by Jeff Sparkes.");
+	action_output("DFT File Transfer Code Copyright October 1995 by Dick Altenbern.");
+	action_output("RPQNAMES Code Copyright 2004, 2005 by Don Russell.");
+	action_output("  Permission to use, copy, modify, and distribute this software and its");
+	action_output("  documentation for any purpose and without fee is hereby granted,");
+	action_output("  provided that the above copyright notice appear in all copies and that");
+	action_output("  both that copyright notice and this permission notice appear in");
+	action_output("  supporting documentation.");
+	action_output(" ");
+	action_output("Copyright 1989 by Georgia Tech Research Corporation, Atlanta, GA 30332.");
+	action_output("  All Rights Reserved.  GTRC hereby grants public use of this software.");
+	action_output("  Derivative works based on this software must incorporate this copyright");
+	action_output("  notice.");
+	action_output(" ");
+	action_output(
 #if defined(_WIN32) /*[*/
-"wc3270"
+	"wc3270"
 #else /*][*/
-"c3270"
+	"c3270"
 #endif /*]*/
-" is distributed in the hope that it will be useful, but WITHOUT ANY\n"
-"WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS\n"
-"FOR A PARTICULAR PURPOSE.  See the file LICENSE for more details.\n"
-"\n");
+	" is distributed in the hope that it will be useful, but WITHOUT ANY");
+	action_output("WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS");
+	action_output("FOR A PARTICULAR PURPOSE.  See the file LICENSE for more details.");
+	action_output(" ");
 }
 
 void
-Show_action(Widget w unused, XEvent *event unused, String *params,
+Show_action(Widget w _is_unused, XEvent *event _is_unused, String *params,
     Cardinal *num_params)
 {
 	action_debug(Show_action, event, params, num_params);
@@ -1020,7 +1060,7 @@ Show_action(Widget w unused, XEvent *event unused, String *params,
 #if defined(X3270_TRACE) /*[*/
 /* Trace([data|keyboard][on[filename]|off]) */
 void
-Trace_action(Widget w unused, XEvent *event unused, String *params,
+Trace_action(Widget w _is_unused, XEvent *event _is_unused, String *params,
     Cardinal *num_params)
 {
 	int tg = 0;
@@ -1083,13 +1123,15 @@ Trace_action(Widget w unused, XEvent *event unused, String *params,
 			appres.trace_file = NewString(params[2]);
 		do_toggle(tg);
 	}
+	if (tracefile_name != NULL)
+		action_output("Trace file is %s", tracefile_name);
 }
 #endif /*]*/
 
 /* Break to the command prompt. */
 void
-Escape_action(Widget w unused, XEvent *event unused, String *params unused,
-    Cardinal *num_params unused)
+Escape_action(Widget w _is_unused, XEvent *event _is_unused, String *params _is_unused,
+    Cardinal *num_params _is_unused)
 {
 	action_debug(Escape_action, event, params, num_params);
 	if (!appres.secure)
