@@ -1,6 +1,5 @@
 /*
- * Modifications Copyright 1993, 1994, 1995, 1996, 2000, 2001, 2002, 2003,
- *    2004, 2005, 2006, 2007 by Paul Mattes.
+ * Modifications Copyright 1993-2008 by Paul Mattes.
  * Original X11 Port Copyright 1990 by Jeff Sparkes.
  *   Permission to use, copy, modify, and distribute this software and its
  *   documentation for any purpose and without fee is hereby granted,
@@ -54,6 +53,10 @@
 #include "trace_dsc.h"
 #include "utilc.h"
 
+#if defined(_WIN32) /*[*/
+#include "winversc.h"
+#endif /*]*/
+
 extern void usage(char *);
 
 #define LAST_ARG	"--"
@@ -61,6 +64,8 @@ extern void usage(char *);
 #if defined(WC3270) /*[*/
 #define PROFILE_SFX	".wc3270"
 #define PROFILE_SFX_LEN	(sizeof(PROFILE_SFX) - 1)
+#define PROFILE_SSFX	".wc3"
+#define PROFILE_SSFX_LEN (sizeof(PROFILE_SSFX) - 1)
 #endif /*]*/
 
 #if defined(C3270) /*[*/
@@ -126,6 +131,11 @@ struct toggle_name toggle_names[N_TOGGLES] = {
 #else /*][*/
 	{ ResAidWait,         -1 },
 #endif /*]*/
+#if defined(C3270) /*[*/
+	{ ResUnderscore,      UNDERSCORE },
+#else /*][*/
+	{ ResUnderscore,      -1 },
+#endif /*]*/
 };
 
 
@@ -211,14 +221,36 @@ parse_command_line(int argc, const char **argv, const char **cl_hostname)
 #if defined(WC3270) /*[*/
 	/* Merge in the profile. */
 	if (*cl_hostname != CN &&
-	    (sl = strlen(*cl_hostname)) > PROFILE_SFX_LEN &&
-	    !strcasecmp(*cl_hostname + sl - PROFILE_SFX_LEN, PROFILE_SFX)) {
+	    (((sl = strlen(*cl_hostname)) > PROFILE_SFX_LEN &&
+	      !strcasecmp(*cl_hostname + sl - PROFILE_SFX_LEN, PROFILE_SFX)) ||
+	     ((sl = strlen(*cl_hostname)) > PROFILE_SSFX_LEN &&
+	      !strcasecmp(*cl_hostname + sl - PROFILE_SSFX_LEN, PROFILE_SSFX)))) {
+
+		const char *pname;
+
 		(void) read_resource_file(*cl_hostname, False);
 		if (appres.hostname == CN) {
 		    Error("Hostname not specified in session file.");
 		}
-		profile_name = NewString(*cl_hostname);
-		profile_name[sl - PROFILE_SFX_LEN] = '\0';
+
+		pname = strrchr(*cl_hostname, '\\');
+		if (pname != CN)
+		    	pname++;
+		else
+		    	pname = *cl_hostname;
+		profile_name = NewString(pname);
+
+		sl = strlen(profile_name);
+		if (sl > PROFILE_SFX_LEN &&
+			!strcasecmp(profile_name + sl - PROFILE_SFX_LEN,
+				PROFILE_SFX)) {
+			profile_name[sl - PROFILE_SFX_LEN] = '\0';
+		} else if (sl > PROFILE_SSFX_LEN &&
+			!strcasecmp(profile_name + sl - PROFILE_SSFX_LEN,
+				PROFILE_SSFX)) {
+			profile_name[sl - PROFILE_SSFX_LEN] = '\0';
+		}
+
 		*cl_hostname = appres.hostname;
 	}
 #endif /*]*/
@@ -329,7 +361,7 @@ parse_options(int *argcp, const char **argv)
 		const char *name;
 		enum {
 		    OPT_BOOLEAN, OPT_STRING, OPT_XRM, OPT_SKIP2, OPT_NOP,
-		    OPT_DONE
+		    OPT_INT, OPT_V, OPT_DONE
 		} type;
 		Boolean flag;
 		const char *res_name;
@@ -359,9 +391,6 @@ parse_options(int *argcp, const char **argv)
     { OptHostsFile,OPT_STRING,  False, ResHostsFile, offset(hostsfile) },
 #if defined(C3270) /*[*/
     { OptKeymap,   OPT_STRING,  False, ResKeymap,    offset(key_map) },
-#endif /*]*/
-#if defined(X3270_DBCS) /*[*/
-    { OptLocalEncoding,OPT_STRING,False,ResLocalEncoding,offset(local_encoding) },
 #endif /*]*/
     { OptModel,    OPT_STRING,  False, ResKeymap,    offset(model) },
 #if defined(C3270) && !defined(_WIN32) /*[*/
@@ -392,6 +421,8 @@ parse_options(int *argcp, const char **argv)
     { OptTraceFile,OPT_STRING,  False, ResTraceFile, offset(trace_file) },
     { OptTraceFileSize,OPT_STRING,False,ResTraceFileSize,offset(trace_file_size) },
 #endif /*]*/
+    { OptV,        OPT_V,	False, NULL,	     NULL },
+    { OptVersion,  OPT_V,	False, NULL,	     NULL },
     { "-xrm",      OPT_XRM,     False, NULL,         NULL },
     { LAST_ARG,    OPT_DONE,    False, NULL,         NULL },
     { CN,          OPT_SKIP2,   False, NULL,         NULL }
@@ -430,7 +461,16 @@ parse_options(int *argcp, const char **argv)
 	appres.model = "4";
 	appres.hostsfile = CN;
 	appres.port = "telnet";
+
+#if !defined(_WIN32) /*[*/
 	appres.charset = "bracket";
+#else /*][*/
+	if (is_nt)
+		appres.charset = "bracket";
+	else
+		appres.charset = "bracket437";
+#endif /*]*/
+
 	appres.termname = CN;
 	appres.macros = CN;
 #if defined(X3270_TRACE) && !defined(_WIN32) /*[*/
@@ -444,6 +484,9 @@ parse_options(int *argcp, const char **argv)
 	appres.meta_escape = "auto";
 	appres.curses_keypad = True;
 	appres.cbreak_mode = False;
+#if defined(CURSES_WIDE) /*[*/
+	appres.acs = True;
+#endif /*]*/
 #endif /*]*/
 
 #if defined(X3270_ANSI) /*[*/
@@ -461,6 +504,7 @@ parse_options(int *argcp, const char **argv)
 #endif /*]*/
 
 	appres.unlock_delay = True;
+	appres.unlock_delay_ms = 350;
 
 #if defined(X3270_FT) /*[*/
 	appres.dft_buffer_size = DFT_BUF;
@@ -472,6 +516,9 @@ parse_options(int *argcp, const char **argv)
 #if defined(X3270_SCRIPT) || defined(TCL3270) /*[*/
 	appres.toggle[AID_WAIT].value = True;
 #endif /*]*/
+#if defined(C3270) && defined(_WIN32) /*[*/
+	appres.toggle[UNDERSCORE].value = True;
+#endif /*]*/
 
 #if defined(C3270) && defined(X3270_SCRIPT) /*[*/
 	appres.plugin_command = "x3270hist.pl";
@@ -522,6 +569,17 @@ parse_options(int *argcp, const char **argv)
 			break;
 		    case OPT_NOP:
 			break;
+		    case OPT_INT:
+			if (i == *argcp - 1)	/* missing arg */
+				continue;
+			*(int *)opts[j].aoff = atoi(argv[++i]);
+			if (opts[j].res_name != CN)
+				add_resource(NewString(opts[j].name),
+					     NewString(argv[i]));
+			break;
+		    case OPT_V:
+			dump_version();
+			break;
 		    case OPT_DONE:
 			while (i < *argcp)
 				argv_out[argc_out++] = argv[i++];
@@ -532,7 +590,7 @@ parse_options(int *argcp, const char **argv)
 	argv_out[argc_out] = CN;
 	(void) memcpy((char *)argv, (char *)argv_out,
 	    (argc_out + 1) * sizeof(char *));
-	Free(argv_out);
+	Free((char *)argv_out);
 
 #if defined(X3270_TRACE) /*[*/
 	/* One isn't very useful without the other. */
@@ -585,7 +643,7 @@ parse_set_clear(int *argcp, const char **argv)
 	argv_out[argc_out] = CN;
 	(void) memcpy((char *)argv, (char *)argv_out,
 	    (argc_out + 1) * sizeof(char *));
-	Free(argv_out);
+	Free((char *)argv_out);
 }
 
 /*
@@ -678,6 +736,9 @@ static struct {
 #endif /*]*/
 	{ ResCharset,	offset(charset),	XRM_STRING },
 	{ ResColor8,	offset(color8),		XRM_BOOLEAN },
+#if defined(TCL3270) /*[*/
+	{ ResCommandTimeout, offset(command_timeout), XRM_INT },
+#endif /*]*/
 	{ ResConfDir,	offset(conf_dir),	XRM_STRING },
 #if defined(C3270) /*[*/
 	{ ResDefScreen,	offset(defscreen),	XRM_STRING },
@@ -688,7 +749,6 @@ static struct {
 #endif /*]*/
 	{ ResExtended,	offset(extended),	XRM_BOOLEAN },
 #if defined(X3270_FT) /*[*/
-	{ ResFtCommand,	offset(ft_command),	XRM_STRING },
 	{ ResDftBufferSize,offset(dft_buffer_size),XRM_INT },
 #endif /*]*/
 #if defined(WC3270) /*[*/
@@ -704,7 +764,7 @@ static struct {
 #if defined(X3270_SCRIPT) /*[*/
 	{ ResPluginCommand, offset(plugin_command), XRM_STRING },
 #endif /*]*/
-#if defined(C3270) && defined(X3270_SCRIPT) /*[*/
+#if defined(C3270) /*[*/
 	{ ResIdleCommand,offset(idle_command),	XRM_STRING },
 	{ ResIdleCommandEnabled,offset(idle_command_enabled),	XRM_BOOLEAN },
 	{ ResIdleTimeout,offset(idle_timeout),	XRM_STRING },
@@ -714,6 +774,9 @@ static struct {
 	{ ResMetaEscape,offset(meta_escape),	XRM_STRING },
 	{ ResCursesKeypad,offset(curses_keypad),XRM_BOOLEAN },
 	{ ResCbreak,	offset(cbreak_mode),	XRM_BOOLEAN },
+#if defined(CURSES_WIDE) /*[*/
+	{ ResAcs,	offset(acs),		XRM_BOOLEAN },
+#endif /*]*/
 #endif /*]*/
 #if defined(X3270_ANSI) /*[*/
 	{ ResKill,	offset(kill),		XRM_STRING },
@@ -756,6 +819,10 @@ static struct {
 #endif /*]*/
 	{ ResTypeahead,	offset(typeahead),	XRM_BOOLEAN },
 	{ ResUnlockDelay,offset(unlock_delay),	XRM_BOOLEAN },
+	{ ResUnlockDelayMs,offset(unlock_delay_ms),XRM_INT },
+#if defined(WC3270) /*[*/
+	{ ResVisualBell,offset(visual_bell),	XRM_BOOLEAN },
+#endif /*]*/
 #if defined(X3270_ANSI) /*[*/
 	{ ResWerase,	offset(werase),		XRM_STRING },
 #endif /*]*/
@@ -810,6 +877,8 @@ parse_xrm(const char *arg, const char *where)
 	enum resource_type type = XRM_STRING;
 #if defined(C3270) /*[*/
 	char *add_buf = CN;
+	char *hide;
+	Boolean arbitrary = False;
 #endif /*]*/
 
 	/* Enforce "-3270." or "-3270*" or "*". */
@@ -876,17 +945,21 @@ parse_xrm(const char *arg, const char *where)
 	if (address == NULL) {
 		if (!strncasecmp(ResKeymap ".", arg + match_len,
 		                 strlen(ResKeymap ".")) ||
-		    !strncasecmp(ResCharset ".", arg + match_len,
-		                 strlen(ResCharset ".")) ||
-		    !strncasecmp(ResDisplayCharset ".", arg + match_len,
-		                 strlen(ResDisplayCharset ".")) ||
-		    !strncasecmp(ResCodepage ".", arg + match_len,
-		                 strlen(ResCodepage ".")) ||
 		    !strncasecmp("host.", arg + match_len, 5) ||
-		    !strncasecmp("printer.", arg + match_len, 8)) {
-			add_buf = Malloc(strlen(s) + 1);
-			address = add_buf;
+		    !strncasecmp("printer.", arg + match_len, 8) ||
+#if defined(_WIN32) /*[*/
+		    !strncasecmp(ResHostColorFor, arg + match_len,
+			    strlen(ResHostColorFor)) ||
+		    !strncasecmp(ResConsoleColorForHostColor, arg + match_len,
+			    strlen(ResConsoleColorForHostColor))
+#else /*][*/
+		    !strncasecmp(ResCursesColorFor, arg + match_len,
+			    strlen(ResCursesColorFor))
+#endif /*]*/
+		    ) {
+			address = &hide;
 			type = XRM_STRING;
+			arbitrary = True;
 		}
 	}
 #endif /*]*/
@@ -963,13 +1036,13 @@ parse_xrm(const char *arg, const char *where)
 
 #if defined(C3270) /*[*/
 	/* Add a new, arbitrarily-named resource. */
-	{
+	if (arbitrary) {
 		char *rsname;
 
 		rsname = Malloc(rnlen + 1);
 		(void) strncpy(rsname, arg + match_len, rnlen);
 		rsname[rnlen] = '\0';
-		add_resource(rsname, NewString(s));
+		add_resource(rsname, hide);
 	}
 #endif /*]*/
 }
@@ -1049,7 +1122,7 @@ read_resource_file(const char *filename, Boolean fatal)
 		/* If this line is a continuation, try again. */
 		if (bsl) {
 			ilen += strlen(buf + ilen);
-			if (ilen >= sizeof(buf) - 1) {
+			if ((unsigned)ilen >= sizeof(buf) - 1) {
 				(void) sprintf(where, "%s:%d: Line too long\n",
 				    filename, lno);
 				Warning(where);
@@ -1179,28 +1252,3 @@ action_output(const char *fmt, ...)
 		macro_output = True;
 	}
 }
-
-#if defined(_WIN32) /*[*/
-
-/* Missing parts for wc3270. */
-#include <windows.h>
-#define SECS_BETWEEN_EPOCHS	11644473600ULL
-#define SECS_TO_100NS		10000000ULL /* 10^7 */
-
-int
-gettimeofday(struct timeval *tv, void *ignored)
-{
-	FILETIME t;
-	ULARGE_INTEGER u;
-
-	GetSystemTimeAsFileTime(&t);
-	memcpy(&u, &t, sizeof(ULARGE_INTEGER));
-
-	/* Isolate seconds and move epochs. */
-	tv->tv_sec = (DWORD)((u.QuadPart / SECS_TO_100NS) -
-			       	SECS_BETWEEN_EPOCHS);
-	tv->tv_usec = (u.QuadPart % SECS_TO_100NS) / 10ULL;
-	return 0;
-}
-
-#endif /*]*/
