@@ -1,6 +1,5 @@
 /*
- * Copyright 1993, 1994, 1995, 1996, 1999, 2000, 2001, 2002, 2003, 2004, 2005,
- *   2006, 2007 by Paul Mattes.
+ * Copyright 1993-2008 by Paul Mattes.
  *  Permission to use, copy, modify, and distribute this software and its
  *  documentation for any purpose and without fee is hereby granted,
  *  provided that the above copyright notice appear in all copies and that
@@ -61,9 +60,18 @@
 #include "tablesc.h"
 #include "telnetc.h"
 #include "trace_dsc.h"
+#include "unicodec.h"
+#include "utf8c.h"
 #include "utilc.h"
 #include "xioc.h"
-#include "widec.h"
+
+#if defined(_WIN32) /*[*/
+#include "windows.h"
+#endif /*]*/
+
+#if defined(_MSC_VER) /*[*/
+#include "Msc/deprecated.h"
+#endif /*]*/
 
 #define ANSI_SAVE_SIZE	4096
 
@@ -107,7 +115,8 @@ typedef struct sms {
 		ST_KEYMAP,	/* keyboard map */
 		ST_IDLE,	/* idle command */
 		ST_CHILD,	/* child process */
-		ST_PEER		/* peer (external) process */
+		ST_PEER,	/* peer (external) process */
+		ST_FILE		/* read commands from file */
 	} type;
 	Boolean	success;
 	Boolean	need_prompt;
@@ -120,6 +129,9 @@ typedef struct sms {
 	unsigned long msec;	/* total accumulated time */
 	FILE   *outfile;
 	int	infd;
+#if defined(_WIN32) /*[*/
+	HANDLE	inhandle;
+#endif /*]*/
 	int	pid;
 	unsigned long expect_id;
 	unsigned long wait_id;
@@ -165,7 +177,8 @@ static int      ansi_save_ix = 0;
 static char    *expect_text = CN;
 static int	expect_len = 0;
 static const char *st_name[] = { "String", "Macro", "Command", "KeymapAction",
-				 "IdleCommand", "ChildScript", "PeerScript" };
+				 "IdleCommand", "ChildScript", "PeerScript",
+				 "File" };
 static enum iaction st_cause[] = { IA_MACRO, IA_MACRO, IA_COMMAND, IA_KEYMAP,
 				 IA_IDLE, IA_MACRO, IA_MACRO };
 #define ST_sNAME(s)	st_name[(int)(s)->type]
@@ -181,6 +194,8 @@ static void sms_pop(Boolean can_exit);
 static void socket_connection(void);
 #endif /*]*/
 static void wait_timed_out(void);
+static void read_from_file(void);
+static sms_t *sms_redirect_to(void);
 
 /* Macro that defines that the keyboard is locked due to user input. */
 #define KBWAIT	(kybdlock & (KL_OIA_LOCKED|KL_OIA_TWAIT|KL_DEFERRED_UNLOCK))
@@ -197,13 +212,6 @@ static void wait_timed_out(void);
     (IN_ANSI && !(kybdlock & KL_AWAITING_FIRST)) \
 )
 
-static unsigned char ldtoasc[] = {
-	0x20, 0x20, 0x20, 0x20, 0x20, 0x20, 0x20, 0xad,
-	0x3e, 0x3c, 0x5b, 0x5d, 0x29, 0x28, 0x7d, 0x7b,
-	0x20, 0x3d, 0x27, 0x22, 0x2f, 0x5c, 0x7c, 0xa6,
-	0x3f, 0x21, 0x24, 0xa2, 0xa3, 0xa5, 0xb6
-};
-
 #if defined(X3270_SCRIPT) && defined(X3270_PLUGIN) /*[*/
 static int plugin_pid = 0;		/* process ID if running, or 0 */
 static int plugin_outpipe = -1;		/* from emulator, to plugin */
@@ -378,7 +386,11 @@ script_enable(void)
 {
 	if (sms->infd >= 0 && stdin_id == 0) {
 		trace_dsn("Enabling input for %s[%d]\n", ST_NAME, sms_depth);
+#if defined(_WIN32) /*[*/
+		stdin_id = AddInput((int)sms->inhandle, script_input);
+#else /*][*/
 		stdin_id = AddInput(sms->infd, script_input);
+#endif /*]*/
 	}
 }
 
@@ -411,6 +423,9 @@ new_sms(enum sms_type type)
 	s->is_login = False;
 	s->outfile = (FILE *)NULL;
 	s->infd = -1;
+#if defined(_WIN32) /*[*/
+	s->inhandle = INVALID_HANDLE_VALUE;
+#endif /*]*/
 	s->pid = -1;
 	s->expect_id = 0L;
 	s->wait_id = 0L;
@@ -517,6 +532,9 @@ sms_pop(Boolean can_exit)
 		(void) fclose(sms->outfile);
 		(void) close(sms->infd);
 	}
+	if (sms->type == ST_FILE) {
+	    	(void) close(sms->infd);
+	}
 
 	/* Cancel any pending timeouts. */
 	if (sms->expect_id != 0L)
@@ -553,9 +571,11 @@ sms_pop(Boolean can_exit)
 		sms->state = SS_KBWAIT;
 		trace_dsn("%s[%d] implicitly paused %s\n",
 			    ST_NAME, sms_depth, sms_state_name[sms->state]);
-	} else if (sms->state == SS_IDLE) {
+	} else if (sms->state == SS_IDLE && sms->type != ST_FILE) {
 		/* The parent needs to be restarted. */
 		script_enable();
+	} else if (sms->type == ST_FILE) {
+	    	read_from_file();
 	}
 }
 
@@ -624,6 +644,9 @@ peer_script_init(void)
 	}
 
 	s->infd = fileno(stdin);
+#if defined(_WIN32) /*[*/
+	s->inhandle = GetStdHandle(STD_INPUT_HANDLE);
+#endif /*]*/
 	s->outfile = stdout;
 	(void) SETLINEBUF(s->outfile);	/* even if it's a pipe */
 
@@ -643,7 +666,7 @@ socket_connection(void)
 {
 	int fd;
 	struct sockaddr_un ssun;
-	size_t len = sizeof(ssun);
+	socklen_t len = sizeof(ssun);
 	sms_t *s;
 
 	/* Accept the connection. */
@@ -670,7 +693,7 @@ socket_connection(void)
 
 /* Clean up the Unix-domain socket. */
 static void
-cleanup_socket(Boolean b unused)
+cleanup_socket(Boolean b _is_unused)
 {
 	char buf[1024];
 
@@ -1166,6 +1189,16 @@ push_string(char *s, Boolean is_login, Boolean is_hex)
 		sms_continue();
 }
 
+/* Push a Source'd file on the stack. */
+static void
+push_file(int fd)
+{
+    	if (!sms_push(ST_FILE))
+	    	return;
+	sms->infd = fd;
+	read_from_file();
+}
+
 /* Set a pending string. */
 void
 ps_set(char *s, Boolean is_hex)
@@ -1296,6 +1329,50 @@ run_script(void)
 	}
 }
 
+/* Read the next command from a file. */
+static void
+read_from_file(void)
+{
+	char *dptr;
+	int len_left = sizeof(sms->msc);
+
+	sms->msc_len = 0;
+	dptr = sms->msc;
+
+	while (len_left) {
+		int nr;
+
+		nr = read(sms->infd, dptr, 1);
+		if (nr < 0) {
+			sms_pop(False);
+			return;
+		}
+		if (nr == 0) {
+		    	if (sms->msc_len == 0) {
+			    	sms_pop(False);
+				return;
+			} else {
+			    	*++dptr = '\0';
+			    	break;
+			}
+		}
+		if (*dptr == '\n') {
+		    	if (sms->msc_len) {
+			    	*dptr = '\0';
+				break;
+			}
+		}
+		dptr++;
+		sms->msc_len++;
+		len_left--;
+	}
+
+	/* Run the command as a macro. */
+	trace_dsn("%s[%d] read '%s'\n", ST_NAME, sms_depth, sms->msc);
+	sms->state = SS_INCOMPLETE;
+	push_macro(sms->dptr, False);
+}
+
 /* Handle an error generated during the execution of a script or macro. */
 void
 sms_error(const char *msg)
@@ -1304,23 +1381,19 @@ sms_error(const char *msg)
 	Boolean is_script = False;
 
 	/* Print the error message. */
-	for (s = sms; s != NULL; s = s->next) {
-		if (sms->type == ST_PEER || sms->type == ST_CHILD) {
-			is_script = True;
-			break;
-		}
-	}
+	s = sms_redirect_to();
+	is_script = (s != NULL);
 	if (is_script) {
 		char c;
 
-		fprintf(sms->outfile, "data: ");
+		fprintf(s->outfile, "data: ");
 		while ((c = *msg++)) {
 			if (c == '\n')
-				(void) putc(' ', sms->outfile);
+				(void) putc(' ', s->outfile);
 			else
-				(void) putc(c, sms->outfile);
+				(void) putc(c, s->outfile);
 		}
-		putc('\n', sms->outfile);
+		putc('\n', s->outfile);
 	} else
 		(void) fprintf(stderr, "%s\n", msg);
 
@@ -1347,6 +1420,7 @@ sms_info(const char *fmt, ...)
 	char msgbuf[4096];
 	char *msg = msgbuf;
 	va_list args;
+	sms_t *s;
 
 	va_start(args, fmt);
 	vsprintf(msgbuf, fmt, args);
@@ -1361,16 +1435,11 @@ sms_info(const char *fmt, ...)
 		else
 			nc = strlen(msg);
 		if (nc || (nl != CN)) {
-			switch (sms->type) {
-			case ST_PEER:
-			case ST_CHILD:
-				(void) fprintf(sms->outfile, "data: %.*s\n",
+		    	if ((s = sms_redirect_to()) != NULL)
+				(void) fprintf(s->outfile, "data: %.*s\n",
 				    nc, msg);
-				break;
-			default:
+			else
 				(void) printf("%.*s\n", nc, msg);
-				break;
-			}
 		}
 		msg = nl + 1;
 	} while (nl);
@@ -1556,6 +1625,9 @@ sms_continue(void)
 			script_enable();
 			run_script();
 			break;
+		    case ST_FILE:
+			read_from_file();
+			break;
 		}
 	}
 
@@ -1583,7 +1655,7 @@ sms_in_macro(void)
 
 static void
 dump_range(int first, int len, Boolean in_ascii, struct ea *buf,
-    int rel_rows unused, int rel_cols)
+    int rel_rows _is_unused, int rel_cols)
 {
 	register int i;
 	Boolean any = False;
@@ -1591,7 +1663,7 @@ dump_range(int first, int len, Boolean in_ascii, struct ea *buf,
 	char *linebuf;
 	char *s;
 
-	linebuf = Malloc(maxCOLS * 3 + 1);
+	linebuf = Malloc(maxCOLS * 4 + 1);
 	s = linebuf;
 
 	/*
@@ -1613,61 +1685,48 @@ dump_range(int first, int len, Boolean in_ascii, struct ea *buf,
 			any = False;
 		}
 		if (in_ascii) {
-			unsigned char c;
+			char mb[16];
+			ucs4_t uc;
+			int j;
+			int xlen;
 
 			if (buf[first + i].fa) {
 				is_zero = FA_IS_ZERO(buf[first + i].fa);
-				c = ' ';
+				s += sprintf(s, " ");
 			} else if (is_zero)
-				c = ' ';
+				s += sprintf(s, " ");
 			else
 #if defined(X3270_DBCS) /*[*/
 			if (IS_LEFT(ctlr_dbcs_state(first + i))) {
-				int len;
-				char mb[16];
-				int j;
 
-				len = dbcs_to_mb(buf[first + i].cc,
-					      buf[first + i + 1].cc,
-					      mb);
-				for (j = 0; j < len; j++) {
+				xlen = ebcdic_to_multibyte(
+					(buf[first + i].cc << 8) |
+					 buf[first + i + 1].cc,
+					mb, sizeof(mb));
+				for (j = 0; j < xlen - 1; j++) {
 					s += sprintf(s, "%c", mb[j]);
 				}
-				continue;
 			} else if (IS_RIGHT(ctlr_dbcs_state(first + i))) {
 				continue;
 			} else
 #endif /*]*/
 			{
-				switch (buf[first + i].cs) {
-				case CS_BASE:
-					c = ebc2asc[buf[first + i].cc];
-					break;
-				case CS_APL:
-				case CS_BASE | CS_GE:
-					c = ge2asc[buf[first + i].cc];
-					if (c < ' ')
-						c = ' ';
-					break;
-				case CS_LINEDRAW:
-					if (buf[first + i].cc <= 0x1f)
-						c = ldtoasc[buf[first + i].cc];
-					else
-						c = ' ';
-					break;
-				default:
-					c = ' ';
-					break;
+				xlen = ebcdic_to_multibyte_x(
+						       buf[first + i].cc,
+						       buf[first + i].cs,
+						       mb, sizeof(mb),
+						       True,
+						       &uc);
+				for (j = 0; j < xlen - 1; j++) {
+					s += sprintf(s, "%c", mb[j]);
 				}
 			}
-			s += sprintf(s, "%c", c ? c : ' ');
 		} else {
 			s += sprintf(s, "%s%02x",
 				i ? " " : "",
 				buf[first + i].cc);
 		}
-		if (!any)
-			any = True;
+		any = True;
 	}
 	if (any) {
 		*s = '\0';
@@ -1762,7 +1821,7 @@ dump_field(Cardinal count, const char *name, Boolean in_ascii)
 }
 
 void
-Ascii_action(Widget w unused, XEvent *event unused, String *params,
+Ascii_action(Widget w _is_unused, XEvent *event _is_unused, String *params,
     Cardinal *num_params)
 {
 	dump_fixed(params, *num_params, action_name(Ascii_action), True,
@@ -1770,14 +1829,14 @@ Ascii_action(Widget w unused, XEvent *event unused, String *params,
 }
 
 void
-AsciiField_action(Widget w unused, XEvent *event unused, String *params unused,
+AsciiField_action(Widget w _is_unused, XEvent *event _is_unused, String *params _is_unused,
     Cardinal *num_params)
 {
 	dump_field(*num_params, action_name(AsciiField_action), True);
 }
 
 void
-Ebcdic_action(Widget w unused, XEvent *event unused, String *params,
+Ebcdic_action(Widget w _is_unused, XEvent *event _is_unused, String *params,
     Cardinal *num_params)
 {
 	dump_fixed(params, *num_params, action_name(Ebcdic_action), False,
@@ -1785,8 +1844,8 @@ Ebcdic_action(Widget w unused, XEvent *event unused, String *params,
 }
 
 void
-EbcdicField_action(Widget w unused, XEvent *event unused,
-    String *params unused, Cardinal *num_params)
+EbcdicField_action(Widget w _is_unused, XEvent *event _is_unused,
+    String *params _is_unused, Cardinal *num_params)
 {
 	dump_field(*num_params, action_name(EbcdicField_action), False);
 }
@@ -1818,7 +1877,6 @@ do_read_buffer(String *params, Cardinal num_params, struct ea *buf, int fd)
 	unsigned char	current_fg = 0x00;
 	unsigned char	current_gr = 0x00;
 	unsigned char	current_cs = 0x00;
-	unsigned char c;
 	Boolean in_ebcdic = False;
 	rpf_t r;
 
@@ -1899,33 +1957,61 @@ do_read_buffer(String *params, Cardinal num_params, struct ea *buf, int fd)
 				else
 					rpf(&r, " %02x", buf[baddr].cc);
 			} else {
-				switch (buf[baddr].cs & CS_MASK) {
-				case CS_BASE:
-				default:
-					if (buf[baddr].cs & CS_GE) {
-						c = ge2asc[buf[baddr].cc];
-						if (c < ' ')
-							c = ' ';
-					} else if (buf[baddr].cc == EBC_null)
-						c = 0;
-					else 
-						c = ebc2asc[buf[baddr].cc];
-					break;
-				case CS_APL:
-					c = ge2asc[buf[baddr].cc];
-					if (c < ' ')
-						c = ' ';
-					break;
-				case CS_LINEDRAW:
-					c = ' ';
-					break;
+			    	Boolean done = False;
+				char mb[16];
+				int j;
+				ucs4_t uc;
 #if defined(X3270_DBCS) /*[*/
-				case CS_DBCS:
-					c = ' '; /* XXX */
+				int len;
+
+				if (IS_LEFT(ctlr_dbcs_state(baddr))) {
+				    	len = ebcdic_to_multibyte(
+						(buf[baddr].cc << 8) |
+						 buf[baddr + 1].cc,
+						mb, sizeof(mb));
+					rpf(&r, " ");
+					for (j = 0; j < len-1; j++)
+						rpf(&r, "%02x", mb[j] & 0xff);
+					done = True;
+				} else if (IS_RIGHT(ctlr_dbcs_state(baddr))) {
+					rpf(&r, " -");
+					done = True;
+				}
+#endif /*]*/
+
+				switch (buf[baddr].cc) {
+				case EBC_null:
+					mb[0] = '\0';
+					break;
+				case EBC_so:
+					mb[0] = 0x0e;
+					mb[1] = '\0';
 					break;
-#endif /*]*/
+				case EBC_si:
+					mb[0] = 0x0f;
+					mb[1] = '\0';
+					break;
+				default:
+					(void) ebcdic_to_multibyte_x(
+							       buf[baddr].cc,
+							       buf[baddr].cs,
+							       mb, sizeof(mb),
+							       False,
+							       &uc);
+				    break;
+				}
+
+				if (!done) {
+					rpf(&r, " ");
+					if (mb[0] == '\0')
+						rpf(&r, "00");
+					else {
+					    	for (j = 0; mb[j]; j++) {
+						    	rpf(&r, "%02x",
+								mb[j] & 0xff);
+						}
+					}
 				}
-				rpf(&r, " %02x", c);
 			}
 		}
 		INC_BA(baddr);
@@ -1942,7 +2028,7 @@ do_read_buffer(String *params, Cardinal num_params, struct ea *buf, int fd)
  * ReadBuffer action.
  */
 void
-ReadBuffer_action(Widget w unused, XEvent *event unused,
+ReadBuffer_action(Widget w _is_unused, XEvent *event _is_unused,
     String *params, Cardinal *num_params)
 {
 	do_read_buffer(params, *num_params, ea_buf, -1);
@@ -2132,7 +2218,7 @@ snap_save(void)
  *      wait for the screen to change, then do a Snap Save
  */
 void
-Snap_action(Widget w unused, XEvent *event unused, String *params,
+Snap_action(Widget w _is_unused, XEvent *event _is_unused, String *params,
     Cardinal *num_params)
 {
 	if (sms == SN || sms->state != SS_RUNNING) {
@@ -2150,7 +2236,7 @@ Snap_action(Widget w unused, XEvent *event unused, String *params,
 	if (!strcasecmp(params[0], action_name(Wait_action))) {
 		long tmo = -1;
 		char *ptr;
-		int maxp = 0;
+		Cardinal maxp = 0;
 
 		if (*num_params > 1 &&
 		    (tmo = strtol(params[1], &ptr, 10)) >= 0 &&
@@ -2279,7 +2365,7 @@ Snap_action(Widget w unused, XEvent *event unused, String *params,
  * Wait for various conditions.
  */
 void
-Wait_action(Widget w unused, XEvent *event unused, String *params,
+Wait_action(Widget w _is_unused, XEvent *event _is_unused, String *params,
     Cardinal *num_params)
 {
 	enum sms_state next_state = SS_WAIT_IFIELD;
@@ -2398,15 +2484,29 @@ sms_host_output(void)
 	}
 }
 
-/* Return whether error pop-ups should be short-circuited. */
+/* Return whether error pop-ups and acition output should be short-circuited. */
+static sms_t *
+sms_redirect_to(void)
+{
+    	sms_t *s;
+
+	for (s = sms; s != SN; s = s->next) {
+	    	if ((s->type == ST_CHILD || s->type == ST_PEER) &&
+		    (s->state == SS_RUNNING ||
+		     s->state == SS_CONNECT_WAIT ||
+		     s->state == SS_WAIT_OUTPUT ||
+		     s->state == SS_SWAIT_OUTPUT ||
+		     s->wait_id != 0L))
+			return s;
+	}
+	return NULL;
+}
+
+/* Return whether error pop-ups and acition output should be short-circuited. */
 Boolean
 sms_redirect(void)
 {
-	return sms != SN &&
-	       (sms->type == ST_CHILD || sms->type == ST_PEER) &&
-	       (sms->state == SS_RUNNING || sms->state == SS_CONNECT_WAIT ||
-		sms->state == SS_WAIT_OUTPUT || sms->state == SS_SWAIT_OUTPUT ||
-		sms->wait_id != 0L);
+    	return sms_redirect_to() != NULL;
 }
 
 #if defined(X3270_MENUS) || defined(C3270) /*[*/
@@ -2557,8 +2657,8 @@ sms_store(unsigned char c)
 
 /* Dump whatever ANSI data has been sent by the host since last called. */
 void
-AnsiText_action(Widget w unused, XEvent *event unused, String *params unused,
-    Cardinal *num_params unused)
+AnsiText_action(Widget w _is_unused, XEvent *event _is_unused, String *params _is_unused,
+    Cardinal *num_params _is_unused)
 {
 	register int i;
 	int ix;
@@ -2597,8 +2697,8 @@ AnsiText_action(Widget w unused, XEvent *event unused, String *params unused,
 
 /* Pause a script. */
 void
-PauseScript_action(Widget w unused, XEvent *event unused, String *params unused,
-    Cardinal *num_params unused)
+PauseScript_action(Widget w _is_unused, XEvent *event _is_unused, String *params _is_unused,
+    Cardinal *num_params _is_unused)
 {
 	if (sms == SN || (sms->type != ST_PEER && sms->type != ST_CHILD)) {
 		popup_an_error("%s can only be called from a script",
@@ -2610,7 +2710,7 @@ PauseScript_action(Widget w unused, XEvent *event unused, String *params unused,
 
 /* Continue a script. */
 void
-ContinueScript_action(Widget w, XEvent *event unused, String *params,
+ContinueScript_action(Widget w, XEvent *event _is_unused, String *params,
     Cardinal *num_params)
 {
 	if (check_usage(ContinueScript_action, *num_params, 1, 1) < 0)
@@ -2637,7 +2737,7 @@ ContinueScript_action(Widget w, XEvent *event unused, String *params,
 
 /* Stop listening to stdin. */
 void
-CloseScript_action(Widget w unused, XEvent *event unused, String *params,
+CloseScript_action(Widget w _is_unused, XEvent *event _is_unused, String *params,
     Cardinal *num_params)
 {
 	if (sms != SN &&
@@ -2662,7 +2762,7 @@ CloseScript_action(Widget w unused, XEvent *event unused, String *params,
 
 /* Execute an arbitrary shell command. */
 void
-Execute_action(Widget w unused, XEvent *event unused, String *params,
+Execute_action(Widget w _is_unused, XEvent *event _is_unused, String *params,
     Cardinal *num_params)
 {
 	if (check_usage(Execute_action, *num_params, 1, 1) < 0)
@@ -2711,7 +2811,7 @@ wait_timed_out(void)
 
 /* Wait for a string from the host (ANSI mode only). */
 void
-Expect_action(Widget w unused, XEvent *event unused, String *params,
+Expect_action(Widget w _is_unused, XEvent *event _is_unused, String *params,
     Cardinal *num_params)
 {
 	int tmo;
@@ -2756,8 +2856,8 @@ static Widget execute_action_shell = (Widget)NULL;
 
 /* Callback for "OK" button on execute action popup */
 static void
-execute_action_callback(Widget w unused, XtPointer client_data,
-    XtPointer call_data unused)
+execute_action_callback(Widget w _is_unused, XtPointer client_data,
+    XtPointer call_data _is_unused)
 {
 	char *text;
 
@@ -2769,8 +2869,8 @@ execute_action_callback(Widget w unused, XtPointer client_data,
 }
 
 void
-execute_action_option(Widget w unused, XtPointer client_data unused,
-    XtPointer call_data unused)
+execute_action_option(Widget w _is_unused, XtPointer client_data _is_unused,
+    XtPointer call_data _is_unused)
 {
 	if (execute_action_shell == NULL)
 		execute_action_shell = create_form_popup("ExecuteAction",
@@ -2784,7 +2884,7 @@ execute_action_option(Widget w unused, XtPointer client_data unused,
 #if defined(X3270_SCRIPT) /*[*/
 /* "Script" action, runs a script as a child process. */
 void
-Script_action(Widget w unused, XEvent *event unused, String *params,
+Script_action(Widget w _is_unused, XEvent *event _is_unused, String *params,
     Cardinal *num_params)
 {
 	int inpipe[2];
@@ -2881,7 +2981,7 @@ Script_action(Widget w unused, XEvent *event unused, String *params,
 
 /* "Macro" action, explicitly invokes a named macro. */
 void
-Macro_action(Widget w unused, XEvent *event unused, String *params,
+Macro_action(Widget w _is_unused, XEvent *event _is_unused, String *params,
     Cardinal *num_params)
 {
 	struct macro_def *m;
@@ -2921,7 +3021,7 @@ cancel_if_idle_command(void)
 #if defined(X3270_PRINTER) /*[*/
 /* "Printer" action, starts or stops a printer session. */
 void
-Printer_action(Widget w unused, XEvent *event unused, String *params,
+Printer_action(Widget w _is_unused, XEvent *event _is_unused, String *params,
     Cardinal *num_params)
 {
 	if (check_usage(Printer_action, *num_params, 1, 2) < 0)
@@ -2956,7 +3056,7 @@ abort_script(void)
 
 /* "Abort" action, stops pending scripts. */
 void
-Abort_action(Widget w unused, XEvent *event unused, String *params,
+Abort_action(Widget w _is_unused, XEvent *event _is_unused, String *params,
     Cardinal *num_params)
 {
 	child_ignore_output();
@@ -2980,9 +3080,8 @@ sms_accumulate_time(struct timeval *t0, struct timeval *t1)
 }
 #endif /*]*/
 
-#if defined(X3270_SCRIPT) /*[*/
 void
-Query_action(Widget w unused, XEvent *event unused, String *params,
+Query_action(Widget w _is_unused, XEvent *event _is_unused, String *params,
     Cardinal *num_params)
 {
 	static struct {
@@ -3024,6 +3123,8 @@ Query_action(Widget w unused, XEvent *event unused, String *params,
 	}
 }
 
+#if defined(X3270_SCRIPT) /*[*/
+
 #if defined(X3270_PLUGIN) /*[*/
 
 /* Plugin facility. */
@@ -3269,7 +3370,7 @@ plugin_start(char *command, char *argv[], Boolean complain)
 }
 
 void
-Plugin_action(Widget w unused, XEvent *event unused, String *params,
+Plugin_action(Widget w _is_unused, XEvent *event _is_unused, String *params,
     Cardinal *num_params)
 {
 	if (*num_params == 0) {
@@ -3435,7 +3536,7 @@ plugin_aid(unsigned char aid)
 
 #else /*][*/
 void
-Plugin_action(Widget w unused, XEvent *event unused, String *params,
+Plugin_action(Widget w _is_unused, XEvent *event _is_unused, String *params,
     Cardinal *num_params)
 {
     /* Do nothing. */
@@ -3448,7 +3549,7 @@ Plugin_action(Widget w unused, XEvent *event unused, String *params,
  * into the trace log.
  */
 void
-Bell_action(Widget w unused, XEvent *event unused, String *params,
+Bell_action(Widget w _is_unused, XEvent *event _is_unused, String *params,
     Cardinal *num_params)
 {
 	ring_bell();
@@ -3456,3 +3557,20 @@ Bell_action(Widget w unused, XEvent *event unused, String *params,
 #endif /*]*/
 
 #endif /*]*/
+
+void
+Source_action(Widget w _is_unused, XEvent *event, String *params,
+    Cardinal *num_params)
+{
+    	int fd;
+
+	action_debug(Source_action, event, params, num_params);
+	if (check_usage(Source_action, *num_params, 1, 1) < 0)
+		return;
+	fd = open(params[0], O_RDONLY);
+	if (fd < 0) {
+	    	popup_an_errno(errno, params[0]);
+		return;
+	}
+	push_file(fd);
+}
