@@ -1,5 +1,5 @@
 /*
- * Copyright 2000, 2001, 2002, 2003, 2004, 2005, 2007 by Paul Mattes.
+ * Copyright 2000-2008 by Paul Mattes.
  *  Permission to use, copy, modify, and distribute this software and its
  *  documentation for any purpose and without fee is hereby granted,
  *  provided that the above copyright notice appear in all copies and that
@@ -23,8 +23,6 @@
  *	    -command "string"
  *		command to use to print (default "lpr", POSIX only)
  *          -charset name
- *          -charset @file
- *          -charset =spec
  *		use the specified character set
  *          -crlf
  *		expand newlines to CR/LF (POSIX only)
@@ -41,13 +39,23 @@
  *	    -printer "printer name"
  *	        printer to use (default is $PRINTER or system default,
  *	        Windows only)
+ *	    -printercp n
+ *	        Code page to use for output (Windows only)
+ *	    -proxy "spec"
+ *	    	proxy specification
  *          -reconnect
  *		keep trying to reconnect
  *	    -trace
  *		trace data stream to a file
  *          -tracedir dir
- *              directory to write trace file in
+ *              directory to write trace file in (POSIX only)
+ *          -trnpre file
+ *          	file of transparent data to send before jobs
+ *          -trnpost file
+ *          	file of transparent data to send after jobs
  *          -v
+ *              display version information and exit
+ *          -V
  *		verbose output about negotiation
  */
 #include <stdio.h>
@@ -60,7 +68,9 @@
 #include <netdb.h>
 #endif /*]*/
 #include <sys/types.h>
+#if !defined(_MSC_VER) /*[*/
 #include <unistd.h>
+#endif /*]*/
 #if !defined(_WIN32) /*[*/
 #include <sys/socket.h>
 #include <netinet/in.h>
@@ -72,13 +82,20 @@
 #endif /*]*/
 #include <time.h>
 #include <signal.h>
+#include <fcntl.h>
 
 #include "globals.h"
+#include "charsetc.h"
 #include "trace_dsc.h"
 #include "ctlrc.h"
+#include "popupsc.h"
+#include "proxyc.h"
+#include "resolverc.h"
 #include "telnetc.h"
+#include "utf8c.h"
 #if defined(_WIN32) /*[*/
 #include "wsc.h"
+#include "windirsc.h"
 #endif /*]*/
 
 #if defined(_IOLBF) /*[*/
@@ -92,13 +109,11 @@
 #endif /*]*/
 
 /* Externals. */
-extern int charset_init(const char *csname);
 extern char *build;
 extern FILE *tracef;
 #if defined(_WIN32) /*[*/
 extern void sockstart(void);
 #endif /*]*/
-extern void sockerr(char *fmt, ...);
 
 /* Globals. */
 char *programname = NULL;	/* program name */
@@ -107,6 +122,7 @@ int ignoreeoj = 0;
 int reconnect = 0;
 #if defined(_WIN32) /*[*/
 int crlf = 1;
+int printercp = 0;
 #else /*][*/
 int crlf = 0;
 #endif /*]*/
@@ -115,6 +131,10 @@ int ffskip = 0;
 int verbose = 0;
 int ssl_host = 0;
 unsigned long eoj_timeout = 0L; /* end of job timeout */
+char *trnpre_data = NULL;
+size_t trnpre_size = 0;
+char *trnpost_data = NULL;
+size_t trnpost_size = 0;
 
 /* User options. */
 #if !defined(_WIN32) /*[*/
@@ -127,24 +147,34 @@ const char *command = "lpr";	/* command to run for printing */
 const char *printer = NULL;	/* printer to use */
 #endif /*]*/
 static int tracing = 0;		/* are we tracing? */
+#if !defined(_WIN32) /*[*/
 static char *tracedir = "/tmp";	/* where we are tracing */
+#endif /*]*/
+char *proxy_spec;		/* proxy specification */
+
+static int proxy_type = 0;
+static char *proxy_host = CN;
+static char *proxy_portname = CN;
+static unsigned short proxy_port = 0;
+
+#if defined(_WIN32) /*[*/
+char appdata[MAX_PATH];
+#endif /* ]*/
 
 void pr3287_exit(int);
+const char *build_options(void);
 
 /* Print a usage message and exit. */
 static void
 usage(void)
 {
 	(void) fprintf(stderr, "usage: %s [options] [lu[,lu...]@]host[:port]\n"
-"Options:\n%s\n%s\n%s\n", programname,
+"Options:\n%s%s%s%s", programname,
 #if !defined(_WIN32) /*[*/
 "  -daemon          become a daemon after connecting\n"
 #endif /*]*/
 "  -assoc <session> associate with a session (TN3270E only)\n"
-"  -charset <name>  use built-in alternate EBCDIC-to-ASCII mappings\n"
-"  -charset @<file> use alternate EBCDIC-to-ASCII mappings from file\n"
-"  -charset =\"<ebc>=<asc> ...\"\n"
-"                   specify alternate EBCDIC-to-ASCII mappings",
+"  -charset <name>  use built-in alternate EBCDIC-to-ASCII mappings\n",
 #if !defined(_WIN32) /*[*/
 "  -command \"<cmd>\" use <cmd> for printing (default \"lpr\")\n"
 #endif /*]*/
@@ -157,16 +187,27 @@ usage(void)
 "  -eojtimeout <seconds>\n"
 "                   time out end of print job\n"
 "  -ffthru          pass through SCS FF orders\n"
-"  -ffskip          skip FF orders at top of page",
+"  -ffskip          skip FF orders at top of page\n",
 "  -ignoreeoj       ignore PRINT-EOJ commands\n"
 #if defined(_WIN32) /*[*/
 "  -printer \"printer name\"\n"
 "                   use specific printer (default is $PRINTER or the system\n"
 "                   default printer)\n"
+"  -printercp <codepage>\n"
+"                   code page for output (default is system ANSI code page)\n"
 #endif /*]*/
+"  -proxy \"<spec>\"\n"
+"                   connect to host via specified proxy\n"
 "  -reconnect       keep trying to reconnect\n"
-"  -trace           trace data stream to /tmp/x3trc.<pid>\n"
-"  -tracedir <dir>  directory to keep trace information in");
+"  -trace           trace data stream to /tmp/x3trc.<pid>\n",
+#if !defined(_WIN32) /*[*/
+"  -tracedir <dir>  directory to keep trace information in\n"
+#endif /*]*/
+"  -trnpre <file>   file of transparent data to send before each job\n"
+"  -trnpost <file>  file of transparent data to send after each job\n"
+"  -v               display version information and exit\n"
+"  -V               log verbose information about connection negotiation\n"
+);
 	pr3287_exit(1);
 }
 
@@ -287,20 +328,48 @@ pr3287_exit(int status)
 	exit(status);
 }
 
+/* Read a transparent data file into memory. */
+static void
+read_trn(char *filename, char **data, size_t *size)
+{
+    	int fd;
+	char buf[1024];
+	int nr;
+
+	if (filename == NULL)
+	    	return;
+	fd = open(filename, O_RDONLY);
+	if (fd < 0) {
+		perror(filename);
+		pr3287_exit(1);
+	}
+	*size = 0;
+	while ((nr = read(fd, buf, sizeof(buf))) > 0) {
+	    	*size += nr;
+	    	*data = realloc(*data, *size);
+		if (*data == NULL) {
+		    	fprintf(stderr, "Out of memory\n");
+			pr3287_exit(1);
+		}
+		memcpy(*data + *size - nr, buf, nr);
+	}
+	if (nr < 0) {
+		perror(filename);
+		pr3287_exit(1);
+	}
+	close(fd);
+}
+
 int
 main(int argc, char *argv[])
 {
 	int i;
 	char *at, *colon;
 	int len;
-	char const *charset = NULL;
+	char *charset = "us";
 	char *lu = NULL;
 	char *host = NULL;
-	const char *port = "telnet";
-#if !defined(AF_INET6) /*[*/
-	struct hostent *he;
-	struct servent *se;
-#endif /*]*/
+	char *port = "telnet";
 	unsigned short p;
 	union {
 		struct sockaddr sa;
@@ -316,6 +385,7 @@ main(int argc, char *argv[])
 #if defined(HAVE_LIBSSL) /*[*/
 	int any_prefixes = False;
 #endif /*]*/
+	char *trnpre = NULL, *trnpost = NULL;
 
 	/* Learn our name. */
 #if defined(_WIN32) /*[*/
@@ -339,6 +409,9 @@ main(int argc, char *argv[])
 	if ((printer = getenv("PRINTER")) == NULL) {
 		printer = ws_default_printer();
 	}
+
+	if (get_dirs(NULL, appdata, "wc3270") < 0)
+	    	exit(1);
 #endif /*]*/
 
 	/* Gather the options. */
@@ -406,13 +479,25 @@ main(int argc, char *argv[])
 			}
 			printer = argv[i + 1];
 			i++;
+		} else if (!strcmp(argv[i], "-printercp")) {
+			if (argc <= i + 1 || !argv[i + 1][0]) {
+				(void) fprintf(stderr,
+				    "Missing value for -printer\n");
+				usage();
+			}
+			printercp = (int)strtoul(argv[i + 1], NULL, 0);
+			i++;
 #endif /*]*/
 		} else if (!strcmp(argv[i], "-reconnect")) {
 			reconnect = 1;
 		} else if (!strcmp(argv[i], "-v")) {
+			printf("%s\n%s\n", build, build_options());
+			exit(0);
+		} else if (!strcmp(argv[i], "-V")) {
 			verbose = 1;
 		} else if (!strcmp(argv[i], "-trace")) {
 			tracing = 1;
+#if !defined(_WIN32) /*[*/
 		} else if (!strcmp(argv[i], "-tracedir")) {
 			if (argc <= i + 1 || !argv[i + 1][0]) {
 				(void) fprintf(stderr,
@@ -421,6 +506,31 @@ main(int argc, char *argv[])
 			}
 			tracedir = argv[i + 1];
 			i++;
+#endif /*]*/
+		} else if (!strcmp(argv[i], "-trnpre")) {
+			if (argc <= i + 1 || !argv[i + 1][0]) {
+				(void) fprintf(stderr,
+				    "Missing value for -trnpre\n");
+				usage();
+			}
+			trnpre = argv[i + 1];
+			i++;
+		} else if (!strcmp(argv[i], "-trnpost")) {
+			if (argc <= i + 1 || !argv[i + 1][0]) {
+				(void) fprintf(stderr,
+				    "Missing value for -trnpost\n");
+				usage();
+			}
+			trnpost = argv[i + 1];
+			i++;
+		} else if (!strcmp(argv[i], "-proxy")) {
+			if (argc <= i + 1 || !argv[i + 1][0]) {
+				(void) fprintf(stderr,
+				    "Missing value for -proxy\n");
+				usage();
+			}
+			proxy_spec = argv[i + 1];
+			i++;
 		} else if (!strcmp(argv[i], "--help")) {
 			usage();
 		} else
@@ -489,32 +599,20 @@ main(int argc, char *argv[])
 		}
 	}
 
-#if !defined(AF_INET6) /*[*/
-	/* Resolve the port number. */
-	se = getservbyname(port, "tcp");
-	if (se != NULL)
-		p = se->s_port;
-	else {
-		unsigned long l;
-		char *ptr;
+#if defined(_WIN32) /*[*/
+	/* Set the printer code page. */
+	if (printercp == 0)
+	    	printercp = GetACP();
+#endif /*]*/
 
-		l = strtoul(port, &ptr, 0);
-		if (ptr == port || *ptr != '\0') {
-			(void) fprintf(stderr, "Unknown port: %s\n", port);
-			pr3287_exit(1);
-		}
-		if (l & ~0xffff) {
-			(void) fprintf(stderr, "Invalid port: %s\n", port);
-			pr3287_exit(1);
-		}
-		p = htons((unsigned short)l);
-	}
-#endif /* AF_INET6 */
-
-	/* Remap the character set. */
+	/* Set up the character set. */
 	if (charset_init(charset) < 0)
 		pr3287_exit(1);
 
+	/* Read in the transparent pre- and post- files. */
+	read_trn(trnpre, &trnpre_data, &trnpre_size);
+	read_trn(trnpost, &trnpost_data, &trnpost_size);
+
 	/* Try opening the trace file, if there is one. */
 	if (tracing) {
 		char tracefile[256];
@@ -522,7 +620,7 @@ main(int argc, char *argv[])
 		int i;
 
 #if defined(_WIN32) /*[*/
-		(void) sprintf(tracefile, "x3trc.%d.txt", getpid());
+		(void) sprintf(tracefile, "%sx3trc.%d.txt", appdata, getpid());
 #else /*][*/
 		(void) sprintf(tracefile, "%s/x3trc.%d", tracedir, getpid());
 #endif /*]*/
@@ -534,8 +632,23 @@ main(int argc, char *argv[])
 		(void) SETLINEBUF(tracef);
 		clk = time((time_t *)0);
 		(void) fprintf(tracef, "Trace started %s", ctime(&clk));
-		(void) fprintf(tracef, " Version %s\n", build);
-		(void) fprintf(tracef, " Options:");
+		(void) fprintf(tracef, " Version: %s\n %s\n", build,
+			       build_options());
+#if !defined(_WIN32) /*[*/
+		(void) fprintf(tracef, " Locale codeset: %s\n", locale_codeset);
+#else /*][*/
+		(void) fprintf(tracef, " ANSI codepage: %d, "
+			       "printer codepage: %d\n", GetACP(), printercp);
+#endif /*]*/
+		(void) fprintf(tracef, " Host codepage: %d",
+			       (int)(cgcsgid & 0xffff));
+#if defined(X3270_DBCS) /*[*/
+		if (dbcs)
+			(void) fprintf(tracef, "+%d",
+				       (int)(cgcsgid_dbcs & 0xffff));
+#endif /*]*/
+		(void) fprintf(tracef, "\n");
+		(void) fprintf(tracef, " Command:");
 		for (i = 0; i < argc; i++) {
 			(void) fprintf(tracef, " %s", argv[i]);
 		}
@@ -574,78 +687,88 @@ main(int argc, char *argv[])
 	(void) signal(SIGPIPE, SIG_IGN);
 #endif /*]*/
 
+	/* Set up the proxy. */
+	if (proxy_spec != CN) {
+	    	proxy_type = proxy_setup(&proxy_host, &proxy_portname);
+		if (proxy_type < 0)
+			pr3287_exit(1);
+	}
+
 	/*
 	 * One-time initialization is now complete.
 	 * (Most) everything beyond this will now be retried, if the -reconnect
 	 * option is in effect.
 	 */
 	for (;;) {
-		/* Resolve the host name. */
-#if defined(AF_INET6) /*[*/
-		struct addrinfo hints, *res;
-		int rc;
+		char errtxt[1024];
 
-		(void) memset(&hints, '\0', sizeof(struct addrinfo));
-		hints.ai_flags = 0;
-		hints.ai_family = PF_UNSPEC;
-		hints.ai_socktype = SOCK_STREAM;
-		hints.ai_protocol = IPPROTO_TCP;
-		rc = getaddrinfo(host, port, &hints, &res);
-		if (rc != 0) {
-			errmsg("%s/%s: %s", host, port, gai_strerror(rc));
-			rc = 1;
-			goto retry;
-		}
-		switch (res->ai_family) {
-		case AF_INET:
-			p = ((struct sockaddr_in *)res->ai_addr)->sin_port;
-			break;
-		case AF_INET6:
-			p = ((struct sockaddr_in6 *)res->ai_addr)->sin6_port;
-			break;
-		default:
-			errmsg("%s: unknown family %d", host, res->ai_family);
-			rc = 1;
-			freeaddrinfo(res);
-			goto retry;
-		}
-		(void) memcpy(&ha.sa, res->ai_addr, res->ai_addrlen);
-		ha_len = res->ai_addrlen;
-		freeaddrinfo(res);
-#else /*][*/
-		(void) memset(&ha, '\0', sizeof(ha));
-		ha.sin.sin_family = AF_INET;
-		ha.sin.sin_port = p;
-		he = gethostbyname(host);
-		if (he != NULL) {
-			(void) memcpy(&ha.sin.sin_addr, he->h_addr_list[0],
-				      he->h_length);
+		/* Resolve the host name. */
+	    	if (proxy_type > 0) {
+		    	unsigned long lport;
+			char *ptr;
+			struct servent *sp;
+
+			if (resolve_host_and_port(proxy_host, proxy_portname,
+				    &proxy_port, &ha.sa, &ha_len, errtxt,
+				    sizeof(errtxt)) < 0) {
+			    popup_an_error("%s/%s: %s", proxy_host,
+				    proxy_portname, errtxt);
+			    rc = 1;
+			    goto retry;
+			}
+
+			lport = strtoul(port, &ptr, 0);
+			if (ptr == port || *ptr != '\0' || lport == 0L ||
+				    lport & ~0xffff) {
+				if (!(sp = getservbyname(port, "tcp"))) {
+					popup_an_error("Unknown port number "
+						"or service: %s", port);
+					rc = 1;
+					goto retry;
+				}
+				p = ntohs(sp->s_port);
+			} else
+				p = (unsigned short)lport;
 		} else {
-			ha.sin.sin_addr.s_addr = inet_addr(host);
-			if (ha.sin.sin_addr.s_addr == INADDR_NONE) {
-				errmsg("Unknown host: %s", host);
-				rc = 1;
-				goto retry;
+			if (resolve_host_and_port(host, port, &p, &ha.sa,
+				    &ha_len, errtxt, sizeof(errtxt)) < 0) {
+			    popup_an_error("%s/%s: %s", host, port, errtxt);
+			    rc = 1;
+			    goto retry;
 			}
 		}
-#endif /*]*/
 
 		/* Connect to the host. */
 		s = socket(ha.sa.sa_family, SOCK_STREAM, 0);
 		if (s < 0) {
-			sockerr("socket");
+			popup_a_sockerr("socket");
 			pr3287_exit(1);
 		}
+
 		if (connect(s, &ha.sa, ha_len) < 0) {
-			sockerr("%s", host);
+			popup_a_sockerr("%s", (proxy_type > 0)? proxy_host:
+								host);
 			rc = 1;
 			goto retry;
 		}
 
+		if (proxy_type > 0) {
+		    	/* Connect to the host through the proxy. */
+		    	if (verbose) {
+			    	(void) fprintf(stderr, "Connected to proxy "
+					       "server %s, port %u\n",
+					       proxy_host, proxy_port);
+			}
+		    	if (proxy_negotiate(proxy_type, s, host, p) < 0) {
+				rc = 1;
+				goto retry;
+			}
+		}
+
 		/* Say hello. */
 		if (verbose) {
 			(void) fprintf(stderr, "Connected to %s, port %u%s\n",
-			    host, ntohs(p),
+			    host, p,
 			    ssl_host? " via SSL": "");
 			if (assoc != NULL)
 				(void) fprintf(stderr, "Associating with LU "
@@ -669,7 +792,7 @@ main(int argc, char *argv[])
 
 		/* Report sudden success. */
 		if (report_success) {
-			errmsg("Connected to %s, port %u", host, ntohs(p));
+			errmsg("Connected to %s, port %u", host, p);
 			report_success = 0;
 		}
 
@@ -750,3 +873,25 @@ popup_an_errno(int err, const char *fmt, ...)
 	}
 	va_end(args);
 }
+
+const char *
+build_options(void)
+{
+    	return "Options:"
+#if defined(X3270_DBCS) /*[*/
+	    " --enable-dbcs"
+#else /*][*/
+	    " --disable-dbcs"
+#endif /*]*/
+#if defined(HAVE_LIBSSL) /*[*/
+	    " --with-ssl"
+#else /*][*/
+	    " --without-ssl"
+#endif /*]*/
+#if defined(USE_ICONV) /*[*/
+	    " --with-iconv"
+#else /*][*/
+	    " --without-iconv"
+#endif /*]*/
+	    ;
+}
