@@ -1,6 +1,5 @@
 /*
- * Copyright 1994, 1995, 1999, 2000, 2001, 2002, 2004, 2005, 2007 by Paul
- *   Mattes.
+ * Copyright 1994-2008 by Paul Mattes.
  *  Permission to use, copy, modify, and distribute this software and its
  *  documentation for any purpose and without fee is hereby granted,
  *  provided that the above copyright notice appear in all copies and that
@@ -38,25 +37,37 @@
 #include "resources.h"
 
 #include "actionsc.h"
+#include "charsetc.h"
 #include "popupsc.h"
 #include "printc.h"
+#include "unicodec.h"
+#include "utf8c.h"
 #include "utilc.h"
-#if defined(X3270_DBCS) /*[*/
-#include "widec.h"
-#endif /*]*/
+
 #if defined(_WIN32) /*[*/
+#include <windows.h>
 #include <io.h>
 #include <fcntl.h>
 #include <sys/stat.h>
 #endif /*]*/
 
+#if defined(_MSC_VER) /*[*/
+#include "Msc/deprecated.h"
+#endif /*]*/
+
+/* Globals */
+#if defined(X3270_DISPLAY) /*[*/
+char *print_text_command = NULL;
+Boolean ptc_changed = FALSE;
+#endif /*]*/
+
 /* Statics */
 
 #if defined(X3270_DISPLAY) /*[*/
 static Widget print_text_shell = (Widget)NULL;
 static Widget save_text_shell = (Widget)NULL;
 static Widget print_window_shell = (Widget)NULL;
-static char *print_window_command = CN;
+char *print_window_command = CN;
 #endif /*]*/
 
 
@@ -118,14 +129,16 @@ html_color(int color)
  * Print the ASCIIfied contents of the screen onto a stream.
  * Returns True if anything printed, False otherwise.
  * 
- * If 'use_html' is True, then HTML is generated, which preserves colors, but
- * little else (for now).
+ * 'ptype' can specify:
+ *  P_TEXT: Ordinary text
+ *  P_HTML: HTML
+ *  P_RTF: Windows rich text
  */
 Boolean
-fprint_screen(FILE *f, Boolean even_if_empty, Boolean use_html)
+fprint_screen(FILE *f, Boolean even_if_empty, ptype_t ptype)
 {
 	register int i;
-	char c;
+	unsigned long uc;
 	int ns = 0;
 	int nr = 0;
 	Boolean any = False;
@@ -134,7 +147,7 @@ fprint_screen(FILE *f, Boolean even_if_empty, Boolean use_html)
 	int fa_color, current_color;
 	Bool fa_high, current_high;
 
-	if (use_html) {
+	if (ptype != P_TEXT) {
 		even_if_empty = True;
 	}
 
@@ -150,18 +163,28 @@ fprint_screen(FILE *f, Boolean even_if_empty, Boolean use_html)
 		fa_high = FA_IS_HIGH(fa);
 	current_high = fa_high;
 
-	for (i = 0; i < ROWS*COLS; i++) {
-#if defined(X3270_DBCS) /*[*/
-		char mb[16];
-		Boolean is_dbcs = False;
+#if defined(_WIN32) /*[*/
+	if (ptype == P_RTF) {
+		fprintf(f, "{\\rtf1\\ansi\\ansicpg%u\\deff0\\deflang1033{\\fonttbl{\\f0\\fmodern\\fprq1\\fcharset0 Courier New;}}\n"
+			    "\\viewkind4\\uc1\\pard\\f0\\fs16 ",
+			    GetACP());
+		if (current_high)
+		    	fprintf(f, "\\b ");
+	}
 #endif /*]*/
 
+	for (i = 0; i < ROWS*COLS; i++) {
+		char mb[16];
+		int nmb;
+
+		uc = 0;
+
 		if (i && !(i % COLS)) {
 			nr++;
 			ns = 0;
 		}
 		if (ea_buf[i].fa) {
-			c = ' ';
+			uc = ' ';
 			fa = ea_buf[i].fa;
 			if (ea_buf[i].fg)
 				fa_color = ea_buf[i].fg & 0x0f;
@@ -172,42 +195,90 @@ fprint_screen(FILE *f, Boolean even_if_empty, Boolean use_html)
 			else
 				fa_high = FA_IS_HIGH(fa);
 		}
-		if (FA_IS_ZERO(fa))
-			c = ' ';
+		if (FA_IS_ZERO(fa)) {
+#if defined(X3270_DBCS) /*[*/
+			if (ctlr_dbcs_state(i) == DBCS_LEFT)
+			    	uc = 0x3000;
+			else
+#endif /*]*/
+				uc = ' ';
+		} else {
+		    	/* Convert EBCDIC to Unicode. */
 #if defined(X3270_DBCS) /*[*/
-		else {
-			/* XXX: DBCS/html interactions are not done */
 			switch (ctlr_dbcs_state(i)) {
 			case DBCS_NONE:
 			case DBCS_SB:
-				c = ebc2asc[ea_buf[i].cc];
+			    	uc = ebcdic_to_unicode(ea_buf[i].cc,
+					ea_buf[i].cs, False);
+				if (uc == 0)
+				    	uc = ' ';
 				break;
 			case DBCS_LEFT:
-				dbcs_to_mb(ea_buf[i].cc, ea_buf[i + 1].cc, mb);
-				is_dbcs = True;
-				c = 'x';
+				uc = ebcdic_to_unicode(
+					(ea_buf[i].cc << 8) |
+					 ea_buf[i + 1].cc,
+					CS_BASE, False);
+				if (uc == 0)
+				    	uc = 0x3000;
 				break;
+			case DBCS_RIGHT:
+				/* skip altogether, we took care of it above */
+				continue;
 			default:
-				c = ' ';
+				uc = ' ';
 				break;
 			}
-		}
 #else /*][*/
-		else
-			c = ebc2asc[ea_buf[i].cc];
+			uc = ebcdic_to_unicode(ea_buf[i].cc, ea_buf[i].cs,
+				False);
+			if (uc == 0)
+				uc = ' ';
 #endif /*]*/
-		if (c == ' ')
+		}
+
+		/* Translate to a type-specific format and write it out. */
+		if (uc == ' ')
 			ns++;
+#if defined(X3270_DBCS) /*[*/
+		else if (uc == 0x3000)
+		    	ns += 2;
+#endif /*]*/
 		else {
 			while (nr) {
+#if defined(_WIN32) /*[*/
+			    	if (ptype == P_RTF)
+				    	fprintf(f, "\\par");
+#endif /*]*/
 				(void) fputc('\n', f);
 				nr--;
 			}
 			while (ns) {
-				(void) fputc(' ', f);
+#if defined(_WIN32) /*[*/
+			    	if (ptype == P_RTF)
+				    	fprintf(f, "\\~");
+				else
+#endif /*]*/
+					(void) fputc(' ', f);
 				ns--;
 			}
-			if (use_html) {
+#if defined(_WIN32) /*[*/
+			if (ptype == P_RTF) {
+				Bool high;
+
+				if (ea_buf[i].gr & GR_INTENSIFY)
+					high = True;
+				else
+					high = fa_high;
+				if (high != current_high) {
+					if (high)
+						fprintf(f, "\\b ");
+					else
+						fprintf(f, "\\b0 ");
+					current_high = high;
+				}
+			}
+#endif /*]*/
+			if (ptype == P_HTML) {
 				int color;
 				Bool high;
 
@@ -235,7 +306,10 @@ fprint_screen(FILE *f, Boolean even_if_empty, Boolean use_html)
 					current_high = high;
 				}
 				if (!any) {
-					fprintf(f, "<html>\n "
+					fprintf(f, "<html>\n"
+						   "<head>\n"
+						   " <meta http-equiv=\"Content-Type\" content=\"text/html; charset=UTF-8\">\n"
+						   "</head>\n"
 						   " <body>\n"
 						   "  <table border=0>"
 						   "<tr bgcolor=black><td>"
@@ -245,18 +319,44 @@ fprint_screen(FILE *f, Boolean even_if_empty, Boolean use_html)
 				}
 			}
 			any = True;
-#if defined(X3270_DBCS) /*[*/
-			if (is_dbcs) {
-				(void) fputs(mb, f);
-				i++;
-			}
-			else
+#if defined(_WIN32) /*[*/
+			if (ptype == P_RTF) {
+				if (uc & ~0x7f) {
+					fprintf(f, "\\u%ld?", uc);
+				} else {
+					nmb = unicode_to_multibyte(uc,
+						mb, sizeof(mb));
+					if (mb[0] == '\\' ||
+						mb[0] == '{' ||
+						mb[0] == '}')
+						fprintf(f, "\\%c",
+							mb[0]);
+					else if (mb[0] == '-')
+						fprintf(f, "\\_");
+					else if (mb[0] == ' ')
+						fprintf(f, "\\~");
+					else
+						fputc(mb[0], f);
+				}
+			} else
 #endif /*]*/
-			{
-				if (use_html && c == '<')
+			if (ptype == P_HTML) {
+				if (uc == '<')
 					fprintf(f, "&lt;");
-				else
-					(void) fputc(c, f);
+				else {
+					nmb = unicode_to_utf8(uc, mb);
+					{
+					    int k;
+
+					    for (k = 0; k < nmb; k++) {
+						fputc(mb[k], f);
+					    }
+					}
+				}
+			} else {
+				nmb = unicode_to_multibyte(uc,
+					mb, sizeof(mb));
+				(void) fputs(mb, f);
 			}
 		}
 	}
@@ -264,10 +364,19 @@ fprint_screen(FILE *f, Boolean even_if_empty, Boolean use_html)
 	if (!any && !even_if_empty)
 		return False;
 	while (nr) {
+#if defined(_WIN32) /*[*/
+	    	if (ptype == P_RTF)
+		    	fprintf(f, "\\par");
+#endif /*]*/
 		(void) fputc('\n', f);
 		nr--;
 	}
-	if (use_html && any) {
+#if defined(_WIN32) /*[*/
+	if (ptype == P_RTF) {
+	    	fprintf(f, "\n}\n%c", 0);
+	}
+#endif /*]*/
+	if ((ptype == P_HTML) && any) {
 		fprintf(f, "%s</font></pre></td></tr>\n"
 		           "  </table>\n"
 			   " </body>\n"
@@ -277,11 +386,12 @@ fprint_screen(FILE *f, Boolean even_if_empty, Boolean use_html)
 	return True;
 }
 
+#if !defined(_WIN32) /*[*/
 /* Termination code for print text process. */
 static void
 print_text_done(FILE *f, Boolean do_popdown
 #if defined(X3270_DISPLAY) /*[*/
-					    unused
+					    _is_unused
 #endif /*]*/
 						  )
 {
@@ -301,12 +411,13 @@ print_text_done(FILE *f, Boolean do_popdown
 	}
 
 }
+#endif /*]*/
 
 #if defined(X3270_DISPLAY) /*[*/
 /* Callback for "OK" button on the print text popup. */
 static void
-print_text_callback(Widget w unused, XtPointer client_data,
-    XtPointer call_data unused)
+print_text_callback(Widget w _is_unused, XtPointer client_data,
+    XtPointer call_data _is_unused)
 {
 	char *filter;
 	FILE *f;
@@ -320,14 +431,19 @@ print_text_callback(Widget w unused, XtPointer client_data,
 		popup_an_errno(errno, "popen(%s)", filter);
 		return;
 	}
-	(void) fprint_screen(f, True, False);
+	if (print_text_command == NULL ||
+		strcmp(print_text_command, filter)) {
+	    Replace(print_text_command, filter);
+	    ptc_changed = True;
+	}
+	(void) fprint_screen(f, True, P_TEXT);
 	print_text_done(f, True);
 }
 
 /* Callback for "Plain Text" button on save text popup. */
 static void
-save_text_plain_callback(Widget w unused, XtPointer client_data,
-    XtPointer call_data unused)
+save_text_plain_callback(Widget w _is_unused, XtPointer client_data,
+    XtPointer call_data _is_unused)
 {
 	char *filename;
 	FILE *f;
@@ -341,7 +457,7 @@ save_text_plain_callback(Widget w unused, XtPointer client_data,
 		popup_an_errno(errno, "%s", filename);
 		return;
 	}
-	(void) fprint_screen(f, True, False);
+	(void) fprint_screen(f, True, P_TEXT);
 	fclose(f);
 	XtPopdown(save_text_shell);
 	if (appres.do_confirms)
@@ -350,8 +466,8 @@ save_text_plain_callback(Widget w unused, XtPointer client_data,
 
 /* Callback for "HTML" button on save text popup. */
 static void
-save_text_html_callback(Widget w unused, XtPointer client_data,
-    XtPointer call_data unused)
+save_text_html_callback(Widget w _is_unused, XtPointer client_data,
+    XtPointer call_data _is_unused)
 {
 	char *filename;
 	FILE *f;
@@ -365,7 +481,7 @@ save_text_html_callback(Widget w unused, XtPointer client_data,
 		popup_an_errno(errno, "%s", filename);
 		return;
 	}
-	(void) fprint_screen(f, True, True);
+	(void) fprint_screen(f, True, P_HTML);
 	fclose(f);
 	XtPopdown(save_text_shell);
 	if (appres.do_confirms)
@@ -405,17 +521,118 @@ popup_save_text(char *filename)
 }
 #endif /*]*/
 
+#if defined(_WIN32) /*[*/
+/*
+ * A Windows version of something like mkstemp().  Creates a temporary
+ * file in $TEMP, returning its path and an open file descriptor.
+ */
+int
+win_mkstemp(char **path, ptype_t ptype)
+{
+	char *s;
+	int fd;
+
+	s = getenv("TEMP");
+	if (s == NULL)
+		s = getenv("TMP");
+	if (s == NULL)
+		s = "C:";
+	*path = xs_buffer("%s\\x3h%u.%s", s, getpid(),
+			    (ptype == P_RTF)? "rtf": "txt");
+	fd = open(*path, O_CREAT | O_RDWR, S_IREAD | S_IWRITE);
+	if (fd < 0) {
+	    Free(*path);
+	    *path = NULL;
+	}
+	return fd;
+}
+
+/*
+ * Find WORDPAD.EXE.
+ */
+#define PROGRAMFILES "%ProgramFiles%"
+char *
+find_wordpad(void)
+{
+	char data[1024];
+	DWORD dlen;
+	char *slash;
+	static char *wp = NULL;
+	HKEY key;
+
+	if (wp != NULL)
+	    return wp;
+
+	/* Get the shell print command for RTF files. */
+	if (RegOpenKeyEx(HKEY_LOCAL_MACHINE,
+		    "Software\\Classes\\rtffile\\shell\\print\\command",
+		    0,
+		    KEY_READ,
+		    &key) != ERROR_SUCCESS) {
+	    	return NULL;
+	}
+	dlen = sizeof(data);
+    	if (RegQueryValueEx(key,
+		    NULL,
+		    NULL,
+		    NULL,
+		    (LPVOID)data,
+		    &dlen) != ERROR_SUCCESS) {
+		RegCloseKey(key);
+	    	return NULL;
+	}
+	RegCloseKey(key);
+
+	if (data[0] == '"') {
+	    char data2[1024];
+	    char *q2;
+
+	    /* The first token is quoted; that's the path. */
+	    strcpy(data2, data + 1);
+	    q2 = strchr(data2, '"');
+	    if (q2 == NULL) {
+		return NULL;
+	    }
+	    *q2 = '\0';
+	    strcpy(data, data2);
+	} else if ((slash = strchr(data, '/')) != NULL) {
+	    /* Find the "/p". */
+	    *slash = '\0';
+	    if (*(slash - 1) == ' ')
+		*(slash - 1) = '\0';
+	}
+
+	if (!strncasecmp(data, PROGRAMFILES, strlen(PROGRAMFILES))) {
+	    char *pf = getenv("PROGRAMFILES");
+
+	    /* Substitute %ProgramFiles%. */
+	    if (pf == NULL) {
+		return NULL;
+	    }
+	    wp = xs_buffer("%s\\%s", pf, data + strlen(PROGRAMFILES));
+	} else {
+	    wp = NewString(data);
+	}
+	if (GetShortPathName(wp, data, sizeof(data)) != 0) {
+	    Free(wp);
+	    wp = NewString(data);
+	}
+	return wp;
+}
+#endif /*]*/
+
 /* Print or save the contents of the screen as text. */
 void
-PrintText_action(Widget w unused, XEvent *event, String *params,
+PrintText_action(Widget w _is_unused, XEvent *event, String *params,
     Cardinal *num_params)
 {
-	int i;
+	Cardinal i;
 	char *filter = CN;
 	Boolean secure = appres.secure;
-	Boolean use_html = False;
+	ptype_t ptype = P_TEXT;
 	Boolean use_file = False;
 	Boolean use_string = False;
+	char *temp_name = NULL;
 
 	action_debug(PrintText_action, event, params, num_params);
 
@@ -438,12 +655,17 @@ PrintText_action(Widget w unused, XEvent *event, String *params,
 			i++;
 			break;
 		} else if (!strcasecmp(params[i], "html")) {
-			use_html = True;
+			ptype = P_HTML;
 			use_file = True;
+#if defined(_WIN32) /*[*/
+		} else if (!strcasecmp(params[i], "rtf")) {
+			ptype = P_RTF;
+			use_file = True;
+#endif /*]*/
 		} else if (!strcasecmp(params[i], "secure")) {
 			secure = True;
 		} else if (!strcasecmp(params[i], "command")) {
-			if (use_html || use_file) {
+			if ((ptype != P_TEXT) || use_file) {
 				popup_an_error("%s: contradictory options",
 				    action_name(PrintText_action));
 				return;
@@ -466,8 +688,13 @@ PrintText_action(Widget w unused, XEvent *event, String *params,
 	switch (*num_params - i) {
 	case 0:
 		/* Use the default. */
-		if (!use_file)
+		if (!use_file) {
+#if !defined(_WIN32) /*[*/
 			filter = get_resource(ResPrintTextCommand);
+#else /*][*/
+			filter = get_resource(ResPrinterName); /* XXX */
+#endif /*]*/
+		}
 		break;
 	case 1:
 		if (use_string) {
@@ -483,6 +710,12 @@ PrintText_action(Widget w unused, XEvent *event, String *params,
 		return;
 	}
 
+#if defined(_WIN32) /*[*/
+	/* On Windows, use rich text. */
+	if (!use_string && !use_file && ptype != P_HTML)
+	    ptype = P_RTF;
+#endif /*]*/
+
 	if (filter != CN && filter[0] == '@') {
 		/*
 		 * Starting the PrintTextCommand resource value with '@'
@@ -493,7 +726,11 @@ PrintText_action(Widget w unused, XEvent *event, String *params,
 		filter++;
 	}
 	if (!use_file && (filter == CN || !*filter))
+#if !defined(_WIN32) /*[*/
 		filter = "lpr";
+#else /*][*/
+		filter = CN;
+#endif /*]*/
 
 #if defined(X3270_DISPLAY) /*[*/
 	if (secure ||
@@ -508,22 +745,16 @@ PrintText_action(Widget w unused, XEvent *event, String *params,
 		/* Invoked non-interactively. */
 		if (use_file) {
 			if (use_string) {
-				char temp_name[15];
-
 #if defined(_WIN32) /*[*/
-				strcpy(temp_name, "x3hXXXXXX");
-				mktemp(temp_name);
-				fd = _open(temp_name, _O_RDWR,
-					_S_IREAD | _S_IWRITE);
+				fd = win_mkstemp(&temp_name, ptype);
 #else /*][*/
-				strcpy(temp_name, "/tmp/x3hXXXXXX");
+				temp_name = NewString("/tmp/x3hXXXXX");
 				fd = mkstemp(temp_name);
 #endif /*]*/
 				if (fd < 0) {
 					popup_an_errno(errno, "mkstemp");
 					return;
 				}
-				(void) unlink(temp_name);
 				f = fdopen(fd, "w+");
 			} else {
 				if (filter == CN || !*filter) {
@@ -533,8 +764,18 @@ PrintText_action(Widget w unused, XEvent *event, String *params,
 				}
 				f = fopen(filter, "a");
 			}
-		} else
+		} else {
+#if !defined(_WIN32) /*[*/
 			f = popen(filter, "w");
+#else /*][*/
+			fd = win_mkstemp(&temp_name, ptype);
+			if (fd < 0) {
+				popup_an_errno(errno, "mkstemp");
+				return;
+			}
+			f = fdopen(fd, "w+");
+#endif /*]*/
+		}
 		if (f == NULL) {
 			popup_an_errno(errno, "%s: %s",
 					action_name(PrintText_action),
@@ -542,9 +783,13 @@ PrintText_action(Widget w unused, XEvent *event, String *params,
 			if (fd >= 0) {
 				(void) close(fd);
 			}
+			if (temp_name) {
+				unlink(temp_name);
+				Free(temp_name);
+			}
 			return;
 		}
-		(void) fprint_screen(f, True, use_html);
+		(void) fprint_screen(f, True, ptype);
 		if (use_string) {
 			char buf[8192];
 
@@ -554,8 +799,35 @@ PrintText_action(Widget w unused, XEvent *event, String *params,
 		}
 		if (use_file)
 			fclose(f);
-		else
+		else {
+#if !defined(_WIN32) /*[*/
 			print_text_done(f, False);
+#else /*][*/
+			char *wp;
+
+			fclose(f);
+			wp = find_wordpad();
+			if (wp == NULL) {
+				popup_an_error("%s: Can't find WORDPAD.EXE",
+					action_name(PrintText_action));
+			} else {
+				char *cmd;
+
+				if (filter != CN)
+				    cmd = xs_buffer("start /wait %s /pt \"%s\" \"%s\"", wp,
+					    temp_name, filter);
+				else
+				    cmd = xs_buffer("start /wait %s /p \"%s\"", wp,
+					    temp_name);
+				system(cmd);
+				Free(cmd);
+			}
+#endif /*]*/
+		}
+		if (temp_name) {
+		    	unlink(temp_name);
+			Free(temp_name);
+		}
 		return;
 	}
 
@@ -575,12 +847,21 @@ PrintText_action(Widget w unused, XEvent *event, String *params,
 
 /* Callback for Print Text menu option. */
 void
-print_text_option(Widget w, XtPointer client_data unused,
-    XtPointer call_data unused)
+print_text_option(Widget w, XtPointer client_data _is_unused,
+    XtPointer call_data _is_unused)
 {
 	char *filter = get_resource(ResPrintTextCommand);
 	Boolean secure = appres.secure;
-	Boolean use_html = False;
+	ptype_t ptype = P_TEXT;
+
+	if (print_text_command != NULL)
+	    	filter = print_text_command;
+	else {
+	    	filter = get_resource(ResPrintTextCommand);
+		if (filter == NULL || !*filter)
+		    	filter = "lpr";
+		print_text_command = XtNewString(filter);
+	}
 
 	/* Decode the filter. */
 	if (filter != CN && *filter == '@') {
@@ -598,7 +879,7 @@ print_text_option(Widget w, XtPointer client_data unused,
 			popup_an_errno(errno, "popen(%s)", filter);
 			return;
 		}
-		(void) fprint_screen(f, True, use_html);
+		(void) fprint_screen(f, True, ptype);
 		print_text_done(f, False);
 	} else {
 		/* Pop up a dialog to confirm or modify their choice. */
@@ -608,8 +889,8 @@ print_text_option(Widget w, XtPointer client_data unused,
 
 /* Callback for Save Text menu option. */
 void
-save_text_option(Widget w, XtPointer client_data unused,
-    XtPointer call_data unused)
+save_text_option(Widget w, XtPointer client_data _is_unused,
+    XtPointer call_data _is_unused)
 {
 	/* Pop up a dialog to confirm or modify their choice. */
 	popup_save_text(CN);
@@ -649,19 +930,18 @@ print_window_done(int status)
 
 /* Timeout callback for window print. */
 static void
-snap_it(XtPointer closure unused, XtIntervalId *id unused)
+snap_it(XtPointer closure _is_unused, XtIntervalId *id _is_unused)
 {
 	if (!print_window_command)
 		return;
 	XSync(display, 0);
 	print_window_done(system(print_window_command));
-	print_window_command = CN;
 }
 
 /* Callback for "OK" button on print window popup. */
 static void
-print_window_callback(Widget w unused, XtPointer client_data,
-    XtPointer call_data unused)
+print_window_callback(Widget w _is_unused, XtPointer client_data,
+    XtPointer call_data _is_unused)
 {
 	print_window_command = XawDialogGetValueString((Widget)client_data);
 	XtPopdown(print_window_shell);
@@ -671,7 +951,7 @@ print_window_callback(Widget w unused, XtPointer client_data,
 
 /* Print the contents of the screen as a bitmap. */
 void
-PrintWindow_action(Widget w unused, XEvent *event, String *params,
+PrintWindow_action(Widget w _is_unused, XEvent *event, String *params,
     Cardinal *num_params)
 {
 	char *filter = get_resource(ResPrintWindowCommand);
@@ -712,8 +992,8 @@ PrintWindow_action(Widget w unused, XEvent *event, String *params,
 #if defined(X3270_MENUS) /*[*/
 /* Callback for menu Print Window option. */
 void
-print_window_option(Widget w, XtPointer client_data unused,
-    XtPointer call_data unused)
+print_window_option(Widget w, XtPointer client_data _is_unused,
+    XtPointer call_data _is_unused)
 {
 	Cardinal zero = 0;
 
@@ -722,3 +1002,4 @@ print_window_option(Widget w, XtPointer client_data unused,
 #endif /*]*/
 
 #endif /*]*/
+
