@@ -1,5 +1,5 @@
 /*
- * Copyright 2000, 2001, 2002, 2003, 2004, 2005 by Paul Mattes.
+ * Copyright 2000-2008 by Paul Mattes.
  *   Permission to use, copy, modify, and distribute this software and its
  *   documentation for any purpose and without fee is hereby granted,
  *   provided that the above copyright notice appear in all copies and that
@@ -25,16 +25,19 @@
 #include "ctlr.h"
 
 #include "actionsc.h"
+#include "charsetc.h"
 #include "ctlrc.h"
 #include "hostc.h"
 #include "keymapc.h"
 #include "kybdc.h"
 #include "macrosc.h"
+#include "popupsc.h"
 #include "screenc.h"
 #include "tablesc.h"
 #include "trace_dsc.h"
+#include "unicodec.h"
+#include "utf8c.h"
 #include "utilc.h"
-#include "widec.h"
 #include "xioc.h"
 
 #undef COLS
@@ -46,9 +49,13 @@ extern int cCOLS;
 #undef COLOR_YELLOW
 #undef COLOR_BLUE
 #undef COLOR_WHITE
-#if defined(HAVE_NCURSES_H) /*[*/
+#if defined(HAVE_NCURSESW_NCURSES_H) /*[*/
+#include <ncursesw/ncurses.h>
+#elif defined(HAVE_NCURSES_NCURSES_H) /*][*/
+#include <ncurses/ncurses.h>
+#elif defined(HAVE_NCURSES_H) /*][*/
 #include <ncurses.h>
-#else /*][*/ 
+#else /*][*/
 #include <curses.h>
 #endif /*]*/
 
@@ -72,6 +79,12 @@ static int cmap[16] = {
 	COLOR_BLACK,	/* gray */
 	COLOR_WHITE	/* white */
 };
+static int field_colors[4] = {
+	COLOR_GREEN,	/* default */
+	COLOR_RED,		/* intensified */
+	COLOR_BLUE,		/* protected */
+	COLOR_WHITE		/* protected, intensified */
+};
 static int defattr = A_NORMAL;
 static unsigned long input_id;
 
@@ -92,13 +105,28 @@ static SCREEN *cur_screen = NULL;
 static void parse_screen_spec(const char *str, struct screen_spec *spec);
 #endif /*]*/
 
+static struct {
+	char *name;
+	int index;
+} cc_name[] = {
+	{ "black",	COLOR_BLACK },
+	{ "red",	COLOR_RED },
+	{ "green",	COLOR_GREEN },
+	{ "yellow",	COLOR_YELLOW },
+	{ "blue",	COLOR_BLUE },
+	{ "magenta", COLOR_MAGENTA },
+	{ "cyan",	COLOR_CYAN },
+	{ "white",	COLOR_WHITE },
+	{ CN,	0 }
+};
+
 static int status_row = 0;	/* Row to display the status line on */
 static int status_skip = 0;	/* Row to blank above the status line */
 
 static Boolean curses_alt = False;
 
 static void kybd_input(void);
-static void kybd_input2(int k, Boolean derived);
+static void kybd_input2(int k, ucs4_t ucs4, int alt);
 static void draw_oia(void);
 static void status_connect(Boolean ignored);
 static void status_3270_mode(Boolean ignored);
@@ -108,8 +136,10 @@ static int color_from_fa(unsigned char);
 static void screen_init2(void);
 static void set_status_row(int screen_rows, int emulator_rows);
 static Boolean ts_value(const char *s, enum ts *tsp);
-static int linedraw_to_acs(unsigned char c);
-static int apl_to_acs(unsigned char c);
+static void display_linedraw(unsigned char ebc);
+static void display_ge(unsigned char ebc);
+static void init_user_colors(void);
+static void init_user_attribute_colors(void);
 
 /* Initialize the screen. */
 void
@@ -295,12 +325,23 @@ screen_init(void)
 		}
 	}
 
-	/* See about keyboard Meta-key behavior. */
+	/*
+	 * See about keyboard Meta-key behavior.
+	 *
+	 * Note: Formerly, "auto" meant to use the terminfo 'km' capability (if
+	 * set, then disable metaEscape).  But popular terminals like the
+	 * Linux console and xterms are actually configurable, though they have
+	 * fixed terminfo capabilities.  It is harmless to enable metaEscape
+	 * when the terminal supports it, so the default is now 'on'.
+	 *
+	 * Setting the high bit for the Meta key is a pretty achaic idea, IMO,
+	 * so we no loger support it.
+	 */
 	if (!ts_value(appres.meta_escape, &me_mode))
 		(void) fprintf(stderr, "invalid %s value: '%s', "
 		    "assuming 'auto'\n", ResMetaEscape, appres.meta_escape);
 	if (me_mode == TS_AUTO)
-		me_mode = tigetflag("km")? TS_OFF: TS_ON;
+		me_mode = TS_ON;
 
 	/* See about all-bold behavior. */
 	if (appres.all_bold_on)
@@ -313,6 +354,10 @@ screen_init(void)
 	if (ab_mode == TS_ON)
 		defattr |= A_BOLD;
 
+	/* Pull in the user's color mappings. */
+	init_user_colors();
+	init_user_attribute_colors();
+
 	/* Set up the controller. */
 	ctlr_init(-1);
 	ctlr_reinit(-1);
@@ -451,22 +496,55 @@ get_color_pair(int fg, int bg)
 }
 
 /*
- * Map a field attribute to a 3270 color index.
+ * Initialize the user-specified attribute color mappings.
+ */
+static void
+init_user_attribute_color(int *a, const char *resname)
+{
+    	char *r;
+	unsigned long l;
+	char *ptr;
+	int i;
+
+    	if ((r = get_resource(resname)) == CN)
+		return;
+	for (i = 0; cc_name[i].name != CN; i++) {
+	    	if (!strcasecmp(r, cc_name[i].name)) {
+		    	*a = cc_name[i].index;
+			return;
+		}
+	}
+	l = strtoul(r, &ptr, 0);
+	if (ptr == r || *ptr != '\0' || l > 7) {
+	    	xs_warning("Invalid %s value: %s", resname, r);
+	    	return;
+	}
+	*a = (int)l;
+}
+
+static void
+init_user_attribute_colors(void)
+{
+	init_user_attribute_color(&field_colors[0],
+		ResCursesColorForDefault);
+	init_user_attribute_color(&field_colors[1],
+		ResCursesColorForIntensified);
+	init_user_attribute_color(&field_colors[2],
+		ResCursesColorForProtected);
+	init_user_attribute_color(&field_colors[3],
+		ResCursesColorForProtectedIntensified);
+}
+
+/*
+ * Map a field attribute to a curses color index.
  * Applies only to m3270 mode -- does not work for mono.
  */
 static int
-color3270_from_fa(unsigned char fa)
+default_color_from_fa(unsigned char fa)
 {
-	static int field_colors[4] = {
-	    COLOR_GREEN,	/* default */
-	    COLOR_RED,		/* intensified */
-	    COLOR_BLUE,		/* protected */
-	    COLOR_WHITE		/* protected, intensified */
 #	define DEFCOLOR_MAP(f) \
 		((((f) & FA_PROTECT) >> 4) | (((f) & FA_INT_HIGH_SEL) >> 3))
 
-	};
-
 	return field_colors[DEFCOLOR_MAP(fa)];
 }
 
@@ -476,7 +554,7 @@ color_from_fa(unsigned char fa)
 	if (appres.m3279) {
 		int fg;
 
-		fg = color3270_from_fa(fa);
+		fg = default_color_from_fa(fa);
 		return get_color_pair(fg, COLOR_BLACK) |
 		    (((ab_mode == TS_ON) || FA_IS_HIGH(fa))? A_BOLD: A_NORMAL);
 	} else if (!appres.mono) {
@@ -488,6 +566,72 @@ color_from_fa(unsigned char fa)
 	}
 }
 
+/*
+ * Set up the user-specified color mappings.
+ */
+static void
+init_user_color(const char *name, int ix)
+{
+    	char *r;
+	int i;
+	unsigned long l;
+	char *ptr;
+
+	r = get_fresource("%s%s", ResCursesColorForHostColor, name);
+	if (r == CN)
+		r = get_fresource("%s%d", ResCursesColorForHostColor, ix);
+	if (r == CN)
+	    	return;
+
+	for (i = 0; cc_name[i].name != CN; i++) {
+	    	if (!strcasecmp(r, cc_name[i].name)) {
+		    	cmap[ix] = cc_name[i].index;
+			return;
+		}
+	}
+
+	l = strtoul(r, &ptr, 0);
+	if (ptr != r && *ptr == '\0' && l <= 7) {
+	    	cmap[ix] = (int)l;
+		return;
+	}
+
+	xs_warning("Invalid %s value '%s'", ResCursesColorForHostColor, r);
+}
+
+static void
+init_user_colors(void)
+{
+	static struct {
+		char *name;
+		int index;
+	} host_color[] = {
+	    	{ "NeutralBlack",	COLOR_NEUTRAL_BLACK },
+	    	{ "Blue",		COLOR_BLUE },
+	    	{ "Red",		COLOR_RED },
+	    	{ "Pink",		COLOR_PINK },
+	    	{ "Green",		COLOR_GREEN },
+	    	{ "Turquoise",		COLOR_TURQUOISE },
+	    	{ "Yellow",		COLOR_YELLOW },
+	    	{ "NeutralWhite",	COLOR_NEUTRAL_WHITE },
+	    	{ "Black",		COLOR_BLACK },
+	    	{ "DeepBlue",		COLOR_DEEP_BLUE },
+	    	{ "Orange",		COLOR_ORANGE },
+	    	{ "Purple",		COLOR_PURPLE },
+	    	{ "PaleGreen",		COLOR_PALE_GREEN },
+	    	{ "PaleTurquoise",	COLOR_PALE_TURQUOISE },
+	    	{ "Grey",		COLOR_GREY },
+	    	{ "Gray",		COLOR_GREY }, /* alias */
+	    	{ "White",		COLOR_WHITE },
+		{ CN,			0 }
+	};
+	int i;
+
+	for (i = 0; host_color[i].name != CN; i++) {
+	    	init_user_color(host_color[i].name, host_color[i].index);
+	}
+}
+
 /*
  * Find the display attributes for a baddr, fa_addr and fa.
  */
@@ -519,7 +663,7 @@ calc_attrs(int baddr, int fa_addr, int fa)
 		else if (ea_buf[fa_addr].fg)
 			fg = cmap[ea_buf[fa_addr].fg & 0x0f];
 		else
-			fg = color3270_from_fa(fa);
+			fg = default_color_from_fa(fa);
 
 		if (ea_buf[baddr].bg)
 			bg = cmap[ea_buf[baddr].bg & 0x0f];
@@ -555,11 +699,10 @@ calc_attrs(int baddr, int fa_addr, int fa)
 
 /* Display what's in the buffer. */
 void
-screen_disp(Boolean erasing unused)
+screen_disp(Boolean erasing _is_unused)
 {
 	int row, col;
 	int field_attrs;
-	int c;
 	unsigned char fa;
 	extern Boolean screen_alt;
 	struct screen_spec *cur_spec;
@@ -611,6 +754,10 @@ screen_disp(Boolean erasing unused)
 		if (!flipped)
 			move(row, 0);
 		for (col = 0; col < cCOLS; col++) {
+		    	Boolean underlined = False;
+			int attr_mask =
+			    toggled(UNDERSCORE)? (int)~A_UNDERLINE: -1;
+
 			if (flipped)
 				move(row, cCOLS-1 - col);
 			baddr = row*cCOLS+col;
@@ -621,57 +768,75 @@ screen_disp(Boolean erasing unused)
 				(void) attrset(defattr);
 				addch(' ');
 			} else if (FA_IS_ZERO(fa)) {
-				(void) attrset(field_attrs);
+				(void) attrset(field_attrs & attr_mask);
+				if (field_attrs & A_UNDERLINE)
+				    underlined = True;
 				addch(' ');
 			} else {
+				char mb[16];
+				int len;
+
 				if (!(ea_buf[baddr].gr ||
 				      ea_buf[baddr].fg ||
 				      ea_buf[baddr].bg)) {
 
-					(void) attrset(field_attrs);
+					(void) attrset(field_attrs & attr_mask);
+					if (field_attrs & A_UNDERLINE)
+					    underlined = True;
 
 				} else {
 				    	int buf_attrs;
 		
 				    	buf_attrs = calc_attrs(baddr, fa_addr,
 						fa);
-					(void) attrset(buf_attrs);
+					(void) attrset(buf_attrs & attr_mask);
+					if (buf_attrs & A_UNDERLINE)
+					    underlined = True;
 				}
 #if defined(X3270_DBCS) /*[*/
 				d = ctlr_dbcs_state(baddr);
 				if (IS_LEFT(d)) {
 					int xaddr = baddr;
-					char mb[16];
-					int len;
-					int i;
 
 					INC_BA(xaddr);
-					len = dbcs_to_mb(ea_buf[baddr].cc,
-					    ea_buf[xaddr].cc,
-					    mb);
-					for (i = 0; i < len; i++) {
-						addch(mb[i] & 0xff);
-					}
+					len = ebcdic_to_multibyte(
+						(ea_buf[baddr].cc << 8) |
+						 ea_buf[xaddr].cc,
+						mb, sizeof(mb));
+					addstr(mb);
 				} else if (!IS_RIGHT(d)) {
 #endif /*]*/
 					if (ea_buf[baddr].cs == CS_LINEDRAW) {
-						c = linedraw_to_acs(ea_buf[baddr].cc);
-						if (c != -1)
-							addch(c);
-						else
-							addch(' ');
+					    	display_linedraw(
+							ea_buf[baddr].cc);
 					} else if (ea_buf[baddr].cs == CS_APL ||
 						   (ea_buf[baddr].cs & CS_GE)) {
-						c = apl_to_acs(ea_buf[baddr].cc);
-						if (c != -1)
-							addch(c);
-						else
-							addch(' ');
+						display_ge(ea_buf[baddr].cc);
 					} else {
-						if (toggled(MONOCASE))
-							addch(asc2uc[ebc2asc[ea_buf[baddr].cc]]);
+						len = ebcdic_to_multibyte(
+							ea_buf[baddr].cc,
+							mb, sizeof(mb));
+						if (len > 0)
+							len--;
+						if (toggled(UNDERSCORE) &&
+							(len == 1) &&
+							mb[0] == ' ') {
+							mb[0] = '_';
+						}
+						if (toggled(MONOCASE) &&
+							(len == 1) &&
+							!(mb[0] & 0x80) &&
+							islower(mb[0])) {
+							mb[0] = toupper(mb[0]);
+						}
+#if defined(CURSES_WIDE) /*[*/
+						addstr(mb);
+#else /*][*/
+						if (len > 1)
+						    	addch(' ');
 						else
-							addch(ebc2asc[ea_buf[baddr].cc]);
+						    	addch(mb[0] & 0xff);
+#endif /*]*/
 					}
 #if defined(X3270_DBCS) /*[*/
 				}
@@ -700,24 +865,34 @@ escape_timeout(void)
 	    "separately\n");
 	eto = 0L;
 	meta_escape = False;
-	kybd_input2(0x1b, False);
+	kybd_input2(0, 0x1b, 0);
 }
 
 /* Keyboard input. */
 static void
 kybd_input(void)
 {
-	int k;
+	int k = 0;		/* KEY_XXX, or 0 */
+	ucs4_t ucs4 = 0;	/* input character, or 0 */
 	Boolean first = True;
 	static Boolean failed_first = False;
 
 	for (;;) {
-		Boolean derived = False;
+		volatile int alt = 0;
 		char dbuf[128];
+#if defined(CURSES_WIDE) /*[*/
+		wint_t wch;
+		size_t sz;
+#endif /*]*/
 
 		if (isendwin())
 			return;
+		ucs4 = 0;
+#if defined(CURSES_WIDE) /*[*/
+		k = wget_wch(stdscr, &wch);
+#else /*][*/
 		k = wgetch(stdscr);
+#endif /*]*/
 		if (k == ERR) {
 			if (first) {
 				if (failed_first) {
@@ -730,7 +905,55 @@ kybd_input(void)
 		} else {
 			failed_first = False;
 		}
-		trace_event("Key %s (0x%x)\n", decode_key(k, 0, dbuf), k);
+#if !defined(CURSES_WIDE) /*[*/
+		/* Differentiate between KEY_XXX and regular input. */
+		if (!(k & ~0xff)) {
+			char mb[2];
+			int consumed;
+			enum me_fail error;
+
+			/* Convert from local multi-byte to Unicode. */
+			mb[0] = k;
+			mb[1] = '\0';
+			ucs4 = multibyte_to_unicode(mb, 1, &consumed, &error);
+			if (ucs4 == 0) {
+				trace_event("Invalid input char 0x%x\n", k);
+				return;
+			}
+			k = 0;
+		}
+#endif /*]*/
+#if defined(CURSES_WIDE) /*[*/
+		if (k == KEY_CODE_YES)
+			k = (int)wch;	/* KEY_XXX */
+		else {
+			char mbs[16];
+			wchar_t wcs[2];
+
+			k = 0;
+			wcs[0] = wch;
+			wcs[1] = 0;
+			sz = wcstombs(mbs, wcs, sizeof(mbs));
+			if (sz < 0) {
+				trace_event("Invalid input wchar 0x%x\n", wch);
+				return;
+			}
+			if (sz == 1) {
+				ucs4 = mbs[0];
+			} else {
+			    	int consumed;
+				enum me_fail error;
+
+			    	ucs4 = multibyte_to_unicode(mbs, sz, &consumed,
+					&error);
+				if (ucs4 == 0) {
+					trace_event("Unsupported input "
+						"wchar %x\n", wch);
+					return;
+				}
+			}
+		}
+#endif /*]*/
 
 		/* Handle Meta-Escapes. */
 		if (meta_escape) {
@@ -739,32 +962,31 @@ kybd_input(void)
 				eto = 0L;
 			}
 			meta_escape = False;
-			k |= 0x80;
-			derived = True;
-		} else if (me_mode == TS_ON && k == 0x1b) {
+			alt = KM_ALT;
+		} else if (me_mode == TS_ON && ucs4 == 0x1b) {
+			trace_event("Key '%s' (curses key 0x%x, char code 0x%x)\n",
+				decode_key(k, ucs4, alt, dbuf), k, ucs4);
 			eto = AddTimeOut(100L, escape_timeout);
 			trace_event(" waiting to see if Escape is followed by"
 			    " another key\n");
 			meta_escape = True;
 			continue;
 		}
-		kybd_input2(k, derived);
+		trace_event("Key '%s' (curses key 0x%x, char code 0x%x)\n",
+			decode_key(k, ucs4, alt, dbuf), k, ucs4);
+		kybd_input2(k, ucs4, alt);
 		first = False;
 	}
 }
 
 static void
-kybd_input2(int k, Boolean derived)
+kybd_input2(int k, ucs4_t ucs4, int alt)
 {
 	char buf[16];
 	char *action;
-	char dbuf1[128], dbuf2[128];
+	int i;
 
-	if (derived)
-		trace_event(" combining <Key>Escape and %s into %s (0x%x)\n",
-		    decode_key(k & 0x7f, 0, dbuf1),
-		    decode_key(k, KM_META, dbuf2), k);
-	action = lookup_key(k);
+	action = lookup_key(k, ucs4, alt);
 	if (action != CN) {
 		if (strcmp(action, "[ignore]"))
 			push_keymap_action(action);
@@ -774,9 +996,6 @@ kybd_input2(int k, Boolean derived)
 
 	/* These first cases apply to both 3270 and NVT modes. */
 	switch (k) {
-	case 0x1d:
-		action_internal(Escape_action, IA_DEFAULT, CN, CN);
-		return;
 	case KEY_UP:
 		action_internal(Up_action, IA_DEFAULT, CN, CN);
 		return;
@@ -795,72 +1014,86 @@ kybd_input2(int k, Boolean derived)
 	default:
 		break;
 	}
+	switch (ucs4) {
+	case 0x1d:
+		action_internal(Escape_action, IA_DEFAULT, CN, CN);
+		return;
+	}
 
 	/* Then look for 3270-only cases. */
-	if (IN_3270) switch(k) {
-	/* These cases apply only to 3270 mode. */
-	case 0x03:
-		action_internal(Clear_action, IA_DEFAULT, CN, CN);
-		return;
-	case 0x12:
-		action_internal(Reset_action, IA_DEFAULT, CN, CN);
-		return;
-	case 'L' & 0x1f:
-		action_internal(Redraw_action, IA_DEFAULT, CN, CN);
-		return;
-	case '\t':
-		action_internal(Tab_action, IA_DEFAULT, CN, CN);
-		return;
-	case 0177:
-	case KEY_DC:
-		action_internal(Delete_action, IA_DEFAULT, CN, CN);
-		return;
-	case '\b':
-	case KEY_BACKSPACE:
-		action_internal(BackSpace_action, IA_DEFAULT, CN, CN);
-		return;
-	case '\r':
-		action_internal(Enter_action, IA_DEFAULT, CN, CN);
-		return;
-	case '\n':
-		action_internal(Newline_action, IA_DEFAULT, CN, CN);
-		return;
-	case KEY_HOME:
-		action_internal(Home_action, IA_DEFAULT, CN, CN);
-		return;
-	default:
-		break;
+	if (IN_3270) {
+	    	switch(k) {
+		case KEY_DC:
+			action_internal(Delete_action, IA_DEFAULT, CN, CN);
+			return;
+		case KEY_BACKSPACE:
+			action_internal(BackSpace_action, IA_DEFAULT, CN, CN);
+			return;
+		case KEY_HOME:
+			action_internal(Home_action, IA_DEFAULT, CN, CN);
+			return;
+		default:
+			break;
+		}
+	    	switch(ucs4) {
+		case 0x03:
+			action_internal(Clear_action, IA_DEFAULT, CN, CN);
+			return;
+		case 0x12:
+			action_internal(Reset_action, IA_DEFAULT, CN, CN);
+			return;
+		case 'L' & 0x1f:
+			action_internal(Redraw_action, IA_DEFAULT, CN, CN);
+			return;
+		case '\t':
+			action_internal(Tab_action, IA_DEFAULT, CN, CN);
+			return;
+		case 0177:
+			action_internal(Delete_action, IA_DEFAULT, CN, CN);
+			return;
+		case '\b':
+			action_internal(BackSpace_action, IA_DEFAULT, CN, CN);
+			return;
+		case '\r':
+			action_internal(Enter_action, IA_DEFAULT, CN, CN);
+			return;
+		case '\n':
+			action_internal(Newline_action, IA_DEFAULT, CN, CN);
+			return;
+		default:
+			break;
+		}
+
 	}
 
 	/* Do some NVT-only translations. */
-	if (IN_ANSI) switch(k) {
+	if (IN_ANSI) switch (k) {
 	case KEY_DC:
-		k = 0x7f;
+	    	ucs4 = 0x7f;
+		k = 0;
 		break;
 	case KEY_BACKSPACE:
-		k = '\b';
+		ucs4 = '\b';
+		k = 0;
 		break;
 	}
 
 	/* Catch PF keys. */
-	if (k >= KEY_F(1) && k <= KEY_F(24)) {
-		(void) sprintf(buf, "%d", k - KEY_F0);
-		action_internal(PF_action, IA_DEFAULT, buf, CN);
-		return;
+	for (i = 0; i < 63; i++) {
+		if (k == KEY_F(i)) {
+			(void) sprintf(buf, "%d", i);
+			action_internal(PF_action, IA_DEFAULT, buf, CN);
+			return;
+		}
 	}
 
 	/* Then any other 8-bit ASCII character. */
-	if (!(k & ~0xff)) {
-		char ks[6];
+	if (ucs4) {
+		char ks[16];
 		String params[2];
 		Cardinal one;
 
-		if (k >= ' ') {
-			ks[0] = k;
-			ks[1] = '\0';
-		} else {
-			(void) sprintf(ks, "0x%x", k);
-		}
+		sprintf(ks, "U+%04x", ucs4);
 		params[0] = ks;
 		params[1] = CN;
 		one = 1;
@@ -941,7 +1174,13 @@ cursor_move(int baddr)
 }
 
 void
-toggle_monocase(struct toggle *t unused, enum toggle_type tt unused)
+toggle_monocase(struct toggle *t _is_unused, enum toggle_type tt _is_unused)
+{
+	screen_disp(False);
+}
+
+void
+toggle_underscore(struct toggle *t _is_unused, enum toggle_type tt _is_unused)
 {
 	screen_disp(False);
 }
@@ -1071,7 +1310,7 @@ status_connect(Boolean connected)
 }
 
 static void
-status_3270_mode(Boolean ignored unused)
+status_3270_mode(Boolean ignored _is_unused)
 {
 	oia_boxsolid = IN_3270 && !IN_SSCP;
 	if (oia_boxsolid)
@@ -1182,8 +1421,8 @@ draw_oia(void)
 }
 
 void
-Redraw_action(Widget w unused, XEvent *event unused, String *params unused,
-    Cardinal *num_params unused)
+Redraw_action(Widget w _is_unused, XEvent *event _is_unused, String *params _is_unused,
+    Cardinal *num_params _is_unused)
 {
 	if (!escaped) {
 		endwin();
@@ -1411,6 +1650,41 @@ linedraw_to_acs(unsigned char c)
 	}
 }
 
+static void
+display_linedraw(unsigned char ebc)
+{
+    	int c;
+	char mb[16];
+	ucs4_t uc;
+	int len;
+
+#if defined(CURSES_WIDE) /*[*/
+	if (appres.acs)
+#endif /*]*/
+	{
+	    	/* Try UCS first. */
+		c = linedraw_to_acs(ebc);
+		if (c != -1) {
+			addch(c);
+			return;
+		}
+	}
+
+	/* Then try Unicode. */
+	len = ebcdic_to_multibyte_x(ebc, CS_LINEDRAW, mb, sizeof(mb), True,
+		&uc);
+	if (len > 0)
+		len--;
+#if defined(CURSES_WIDE) /*[*/
+	addstr(mb);
+#else /*][*/
+	if (len > 1)
+		addch(mb[0] & 0xff);
+	else
+		addch(' ');
+#endif /*]*/
+}
+
 static int
 apl_to_acs(unsigned char c)
 {
@@ -1487,3 +1761,38 @@ apl_to_acs(unsigned char c)
 		return -1;
 	}
 }
+
+static void
+display_ge(unsigned char ebc)
+{
+    	int c;
+	char mb[16];
+	ucs4_t uc;
+	int len;
+
+#if defined(CURSES_WIDE) /*[*/
+	if (appres.acs)
+#endif /*]*/
+	{
+	    	/* Try UCS first. */
+		c = apl_to_acs(ebc);
+		if (c != -1) {
+			addch(c);
+			return;
+		}
+	}
+
+	/* Then try Unicode. */
+	len = ebcdic_to_multibyte_x(ebc, CS_GE, mb, sizeof(mb), True,
+		&uc);
+	if (len > 0)
+		len--;
+#if defined(CURSES_WIDE) /*[*/
+	addstr(mb);
+#else /*][*/
+	if (len > 1)
+		addch(mb[0] & 0xff);
+	else
+		addch(' ');
+#endif /*]*/
+}
