@@ -1,6 +1,5 @@
 /*
- * Modifications Copyright 1993, 1994, 1995, 1999, 2000, 2001, 2002, 2003,
- *   2004, 2005, 2007 by Paul Mattes.
+ * Modifications Copyright 1993-2008 by Paul Mattes.
  * RPQNAMES modifications Copyright 2004 by Don Russell.
  * Original X11 Port Copyright 1990 by Jeff Sparkes.
  *  Permission to use, copy, modify, and distribute this software and its
@@ -30,7 +29,6 @@
 #if defined(_WIN32) /*[*/
 #include <winsock2.h>
 #include <ws2tcpip.h>
-#undef AF_INET6
 #else /*][*/
 #include <sys/socket.h>
 #include <sys/ioctl.h>
@@ -64,11 +62,14 @@
 #include "macrosc.h"
 #include "popupsc.h"
 #include "proxyc.h"
+#include "resolverc.h"
 #include "statusc.h"
 #include "tablesc.h"
 #include "telnetc.h"
 #include "trace_dsc.h"
+#include "unicodec.h"
 #include "utilc.h"
+#include "w3miscc.h"
 #include "xioc.h"
 
 #if !defined(TELOPT_NAWS) /*[*/
@@ -156,7 +157,7 @@ static char     vlnext;
 static int	tn3270e_negotiated = 0;
 static enum { E_NONE, E_3270, E_NVT, E_SSCP } tn3270e_submode = E_NONE;
 static int	tn3270e_bound = 0;
-static char	plu_name[BIND_PLU_NAME_MAX+1];
+static char	*plu_name = NULL;
 static char	**lus = (char **)NULL;
 static char	**curr_lu = (char **)NULL;
 static char	*try_lu = CN;
@@ -343,7 +344,8 @@ sockstart(void)
 	wVersionRequested = MAKEWORD(2, 2);
  
 	if (WSAStartup(wVersionRequested, &wsaData) != 0) {
-		fprintf(stderr, "WSAStartup failed: %ld\n", GetLastError());
+		fprintf(stderr, "WSAStartup failed: %s\n",
+			win32_strerror(GetLastError()));
 		x3270_exit(1);
 	}
  
@@ -365,25 +367,6 @@ static union {
 socklen_t ha_len = sizeof(haddr);
 
 #if defined(_WIN32) /*[*/
-static char *
-winsock_strerror(int e)
-{
-	static char buffer[4096];
-
-	if (FormatMessage(FORMAT_MESSAGE_FROM_SYSTEM,
-	    NULL,
-	    e,
-	    MAKELANGID(LANG_NEUTRAL, SUBLANG_DEFAULT),
-	    buffer,
-	    sizeof(buffer),
-	    NULL) == 0) {
-	    
-	    sprintf(buffer, "Windows error %d", e);
-	}
-
-	return buffer;
-}
-
 void
 popup_a_sockerr(char *fmt, ...)
 {
@@ -393,7 +376,7 @@ popup_a_sockerr(char *fmt, ...)
 	 va_start(args, fmt);
 	 vsprintf(buffer, fmt, args);
 	 va_end(args);
-	 popup_an_error("%s: %s", buffer, winsock_strerror(socket_errno()));
+	 popup_an_error("%s: %s", buffer, win32_strerror(socket_errno()));
 }
 #else /*][*/
 void
@@ -409,86 +392,6 @@ popup_a_sockerr(char *fmt, ...)
 }
 #endif /*]*/
 
-static int
-resolve_host_and_port(const char *host, char *portname, unsigned short *pport)
-{
-#if defined(AF_INET6) /*[*/
-	struct addrinfo	 hints, *res;
-	int		 rc;
-
-	/* Use getaddrinfo() to resolve the hostname and port together. */
-	(void) memset(&hints, '\0', sizeof(struct addrinfo));
-	hints.ai_flags = 0;
-	hints.ai_family = PF_UNSPEC;
-	hints.ai_socktype = SOCK_STREAM;
-	hints.ai_protocol = IPPROTO_TCP;
-	rc = getaddrinfo(host, portname, &hints, &res);
-	if (rc != 0) {
-		popup_an_error("%s/%s: %s", host, portname,
-				gai_strerror(rc));
-		return -1;
-	}
-	switch (res->ai_family) {
-	case AF_INET:
-		*pport =
-		    ntohs(((struct sockaddr_in *)res->ai_addr)->sin_port);
-		break;
-	case AF_INET6:
-		*pport =
-		    ntohs(((struct sockaddr_in6 *)res->ai_addr)->sin6_port);
-		break;
-	default:
-		popup_an_error("%s: unknown family %d", host,
-			res->ai_family);
-		freeaddrinfo(res);
-		return -1;
-	}
-	(void) memcpy(&haddr.sa, res->ai_addr, res->ai_addrlen);
-	ha_len = res->ai_addrlen;
-	freeaddrinfo(res);
-
-#else /*][*/
-
-	struct hostent	*hp;
-	struct servent	*sp;
-	unsigned short	 port;
-	unsigned long	 lport;
-	char		*ptr;
-
-	/* Get the port number. */
-	lport = strtoul(portname, &ptr, 0);
-	if (ptr == portname || *ptr != '\0' || lport == 0L || lport & ~0xffff) {
-		if (!(sp = getservbyname(portname, "tcp"))) {
-			popup_an_error("Unknown port number or service: %s",
-			    portname);
-			return -1;
-		}
-		port = sp->s_port;
-	} else
-		port = htons((unsigned short)lport);
-	*pport = ntohs(port);
-
-	/* Use gethostbyname() to resolve the hostname. */
-	hp = gethostbyname(host);
-	if (hp == (struct hostent *) 0) {
-		haddr.sin.sin_family = AF_INET;
-		haddr.sin.sin_addr.s_addr = inet_addr(host);
-		if (haddr.sin.sin_addr.s_addr == (unsigned long)-1) {
-			popup_an_error("Unknown host:\n%s", host);
-			return -1;
-		}
-	} else {
-		haddr.sin.sin_family = hp->h_addrtype;
-		(void) memmove(&haddr.sin.sin_addr, hp->h_addr,
-				   hp->h_length);
-	}
-	haddr.sin.sin_port = port;
-	ha_len = sizeof(struct sockaddr_in);
-#endif /*]*/
-
-	return 0;
-}
-
 /*
  * net_connect
  *	Establish a telnet socket to the given host passed as an argument.
@@ -505,6 +408,7 @@ net_connect(const char *host, char *portname, Boolean ls, Boolean *resolving,
 	int			passthru_len = 0;
 	unsigned short		passthru_port = 0;
 	int			on = 1;
+	char			errmsg[1024];
 #if defined(OMTU) /*[*/
 	int			mtu = OMTU;
 #endif /*]*/
@@ -591,7 +495,9 @@ net_connect(const char *host, char *portname, Boolean ls, Boolean *resolving,
 		ha_len = sizeof(struct sockaddr_in);
 	} else if (proxy_type > 0) {
 	    	if (resolve_host_and_port(proxy_host, proxy_portname,
-			    &proxy_port) < 0) {
+			    &proxy_port, &haddr.sa, &ha_len, errmsg,
+			    sizeof(errmsg)) < 0) {
+		    	popup_an_error(errmsg);
 		    	return -1;
 		}
 	} else {
@@ -604,8 +510,11 @@ net_connect(const char *host, char *portname, Boolean ls, Boolean *resolving,
 			local_process = False;
 #endif /*]*/
 			if (resolve_host_and_port(host, portname,
-				    &current_port) < 0)
+				    &current_port, &haddr.sa, &ha_len,
+				    errmsg, sizeof(errmsg)) < 0) {
+			    	popup_an_error(errmsg);
 			    	return -1;
+			}
 #if defined(LOCAL_PROCESS) /*[*/
 		}
 #endif /*]*/
@@ -739,15 +648,15 @@ net_connect(const char *host, char *portname, Boolean ls, Boolean *resolving,
 
 		sock_handle = CreateEvent(NULL, TRUE, FALSE, ename);
 		if (sock_handle == NULL) {
-			fprintf(stderr, "Cannot create socket handle: %ld\n",
-			    GetLastError());
+			fprintf(stderr, "Cannot create socket handle: %s\n",
+			    win32_strerror(GetLastError()));
 			x3270_exit(1);
 		}
 	}
 	if (WSAEventSelect(sock, sock_handle, FD_READ | FD_CONNECT | FD_CLOSE)
 		    != 0) {
-		fprintf(stderr, "WSAEventSelect failed: %ld\n",
-		    GetLastError());
+		fprintf(stderr, "WSAEventSelect failed: %s\n",
+		    win32_strerror(GetLastError()));
 		x3270_exit(1);
 	}
 
@@ -1006,8 +915,8 @@ net_input(void)
 					return;
 				default:
 					fprintf(stderr,
-					    "second connect() failed: %d\n",
-					    err);
+					    "second connect() failed: %s\n",
+					    win32_strerror(err));
 					x3270_exit(1);
 				}
 			}
@@ -1066,7 +975,14 @@ net_input(void)
 				return;
 			}
 #endif /*]*/
-			trace_dsn("RCVD socket error %d\n", errno);
+			trace_dsn("RCVD socket error %d (%s)\n",
+				socket_errno(),
+#if !defined(_WIN32) /*[*/
+				strerror(errno)
+#else /*][*/
+				win32_strerror(GetLastError())
+#endif /*]*/
+				);
 			if (HALF_CONNECTED) {
 				popup_a_sockerr("Connect to %s, port %d",
 				    hostname, current_port);
@@ -1567,7 +1483,7 @@ tn3270e_request(void)
 
 	trace_dsn("SENT %s %s DEVICE-TYPE REQUEST %.*s%s%s "
 		   "%s\n",
-	    cmd(SB), opt(TELOPT_TN3270E), strlen(termtype), tt_out + 5,
+	    cmd(SB), opt(TELOPT_TN3270E), (int)strlen(termtype), tt_out + 5,
 	    (try_lu != CN && *try_lu) ? " CONNECT " : "",
 	    (try_lu != CN && *try_lu) ? try_lu : "",
 	    cmd(SE));
@@ -1874,8 +1790,11 @@ static void
 process_bind(unsigned char *buf, int buflen)
 {
 	int namelen, i;
+	int dest_ix = 0;
 
-	(void) memset(plu_name, '\0', sizeof(plu_name));
+	if (plu_name == CN)
+	    	plu_name = Malloc(mb_max_len(BIND_PLU_NAME_MAX));
+	(void) memset(plu_name, '\0', mb_max_len(BIND_PLU_NAME_MAX));
 
 	/* Make sure it's a BIND. */
 	if (buflen < 1 || buf[0] != BIND_RU) {
@@ -1891,7 +1810,12 @@ process_bind(unsigned char *buf, int buflen)
 	if (namelen > BIND_PLU_NAME_MAX)
 		namelen = BIND_PLU_NAME_MAX;
 	for (i = 0; i < namelen; i++) {
-		plu_name[i] = ebc2asc0[buf[BIND_OFF_PLU_NAME + i]];
+	    	int nx;
+
+		nx = ebcdic_to_multibyte(buf[BIND_OFF_PLU_NAME + i],
+			plu_name + dest_ix, mb_max_len(1));
+		if (nx > 1)
+			dest_ix += nx - 1;
 	}
 }
 #endif /*]*/
@@ -2067,7 +1991,14 @@ net_rawout(unsigned const char *buf, int len)
 				return;
 			}
 #endif /*]*/
-			trace_dsn("RCVD socket error %d\n", errno);
+			trace_dsn("RCVD socket error %d (%s)\n",
+				socket_errno(),
+#if !defined(_WIN32) /*[*/
+				strerror(errno)
+#else /*][*/
+				win32_strerror(GetLastError())
+#endif /*]*/
+				);
 			if (socket_errno() == SE_EPIPE || socket_errno() == SE_ECONNRESET) {
 				host_disconnect(False);
 				return;
@@ -3308,7 +3239,7 @@ client_info_callback(INFO_CONST SSL *s, int where, int ret)
 			    SSL_state_string_long(s));
 		} else if (ret < 0) {
 			unsigned long e;
-			char err_buf[121];
+			char err_buf[1024];
 
 			err_buf[0] = '\n';
 			e = ERR_get_error();
@@ -3316,8 +3247,8 @@ client_info_callback(INFO_CONST SSL *s, int where, int ret)
 				(void) ERR_error_string(e, err_buf + 1);
 #if defined(_WIN32) /*[*/
 			else if (GetLastError() != 0)
-				sprintf(err_buf + 1, "error %ld",
-				    GetLastError());
+				strcpy(err_buf + 1,
+					win32_strerror(GetLastError()));
 #else /*][*/
 			else if (errno != 0)
 				strcpy(err_buf + 1, strerror(errno));
@@ -3401,8 +3332,6 @@ continue_tls(unsigned char *sbbuf, int len)
 
 #endif /*]*/
 
-#if defined(X3270_SCRIPT) || defined(TCL3270) /*[*/
-
 /* Return the current BIND application name, if any. */
 const char *
 net_query_bind_plu_name(void)
@@ -3414,7 +3343,7 @@ net_query_bind_plu_name(void)
 	 */
 	if ((cstate == CONNECTED_TN3270E) &&
 	    (e_funcs & E_OPT(TN3270E_FUNC_BIND_IMAGE)))
-		return plu_name;
+		return plu_name? plu_name: "";
 	else
 		return "";
 #else /*][*/
@@ -3496,8 +3425,6 @@ net_query_host(void)
 		return "";
 }
 
-#endif /*]*/
-
 /* Return the local address for the socket. */
 int
 net_getsockname(void *buf, int *len)
