@@ -1,6 +1,5 @@
 /*
- * Copyright 1993, 1994, 1995, 1999, 2000, 2001, 2002, 2003, 2004, 2005, 2007
- *   by Paul Mattes.
+ * Copyright 1993-2008 by Paul Mattes.
  *  Permission to use, copy, modify, and distribute this software and its
  *  documentation for any purpose and without fee is hereby granted,
  *  provided that the above copyright notice appear in all copies and that
@@ -50,7 +49,13 @@
 #include "tablesc.h"
 #include "telnetc.h"
 #include "trace_dsc.h"
+#include "utf8c.h"
 #include "utilc.h"
+#include "w3miscc.h"
+
+#if defined(_MSC_VER) /*[*/
+#include "Msc/deprecated.h"
+#endif /*]*/
 
 /* Maximum size of a tracefile header. */
 #define MAX_HEADER_SIZE		(10*1024)
@@ -85,8 +90,9 @@ static char    *create_tracefile_header(const char *mode);
 static void	stop_tracing(void);
 
 /* Globals */
-struct timeval  ds_ts;
-Boolean         trace_skipping = False;
+struct timeval   ds_ts;
+Boolean          trace_skipping = False;
+char		*tracefile_name = NULL;
 
 /* display a (row,col) */
 const char *
@@ -103,39 +109,92 @@ rcba(int baddr)
 static char *tdsbuf = CN;
 #define TDS_LEN	75
 
+/*
+ * This function is careful to do line breaks based on wchar_t's, not
+ * bytes, so multi-byte characters are traced properly.
+ * However, it doesn't know that DBCS characters are two columns wide, so it
+ * will get those wrong and break too late.  To get that right, it needs some
+ * sort of function to tell it that a wchar_t is double-width, which we lack at
+ * the moment.
+ *
+ * If wchar_t's are Unicode, it could perhaps use some sort of heuristic based
+ * on which plane the character is in.
+ */
 static void
 trace_ds_s(char *s, Boolean can_break)
 {
-	int len = strlen(s);
+    	int len = strlen(s);
+	int len0 = len + 1;
+	int wlen;
 	Boolean nl = False;
+	wchar_t *w_buf;		/* wchar_t translation of s */
+	wchar_t *w_cur;		/* current wchar_t pointer */
+	wchar_t *w_chunk;	/* transient wchar_t buffer */
+	char *mb_chunk;		/* transient multibyte buffer */
 
 	if (!toggled(DS_TRACE) || tracef == NULL || !len)
 		return;
 
-	if (s && s[len-1] == '\n') {
-		len--;
+	/* Allocate buffers for chunks of output data. */
+	mb_chunk = Malloc(len0);
+	w_chunk = (wchar_t *)Malloc(len0 * sizeof(wchar_t));
+
+	/* Convert the input string to wchar_t's. */
+	w_buf = (wchar_t *)Malloc(len0 * sizeof(wchar_t));
+	wlen = mbstowcs(w_buf, s, len);
+	if (wlen < 0)
+	    Error("trace_ds_s: mbstowcs failed");
+	w_cur = w_buf;
+
+	/* Check for a trailing newline. */
+	if (len && s[len-1] == '\n') {
+		wlen--;
 		nl = True;
 	}
-	if (!can_break && dscnt + len >= 75) {
+
+	if (!can_break && dscnt + wlen >= 75) {
 		wtrace("...\n... ");
 		dscnt = 0;
 	}
-	while (dscnt + len >= 75) {
-		int plen = 75-dscnt;
 
-		wtrace("%.*s ...\n... ", plen, s);
+	while (dscnt + wlen >= 75) {
+		int plen = 75-dscnt;
+		int mblen;
+
+		if (plen) {
+		    memcpy(w_chunk, w_cur, plen * sizeof(wchar_t));
+		    w_chunk[plen] = 0;
+		    mblen = wcstombs(mb_chunk, w_chunk, len0);
+		    if (mblen <= 0)
+			Error("trace_ds_s: wcstombs 1 failed");
+		} else {
+		    mb_chunk[0] = '\0';
+		    mblen = 0;
+		}
+
+		wtrace("%.*s ...\n... ", mblen, mb_chunk);
 		dscnt = 4;
-		s += plen;
-		len -= plen;
+		w_cur += plen;
+		wlen -= plen;
 	}
-	if (len) {
-		wtrace("%.*s", len, s);
-		dscnt += len;
+	if (wlen) {
+		int mblen;
+
+		memcpy(w_chunk, w_cur, wlen * sizeof(wchar_t));
+		w_chunk[wlen] = 0;
+		mblen = wcstombs(mb_chunk, w_chunk, len0);
+		if (mblen <= 0)
+		    Error("trace_ds_s: wcstombs 2 failed");
+		wtrace("%.*s", mblen, mb_chunk);
+		dscnt += wlen;
 	}
 	if (nl) {
 		wtrace("\n");
 		dscnt = 0;
 	}
+	Free(mb_chunk);
+	Free(w_buf);
+	Free(w_chunk);
 }
 
 void
@@ -219,13 +278,19 @@ vwtrace(const char *fmt, va_list args)
 	if (tracef_bufptr != CN) {
 		tracef_bufptr += vsprintf(tracef_bufptr, fmt, args);
 	} else if (tracef != NULL) {
-		int nw;
+		int n2w, nw;
+		char buf[16384];
 
-		nw = vfprintf(tracef, fmt, args);
-		if (nw > 0) {
+		buf[0] = 0;
+		(void) vsnprintf(buf, sizeof(buf), fmt, args);
+		buf[sizeof(buf) - 1] = '\0';
+		n2w = strlen(buf);
+
+		nw = fwrite(buf, n2w, 1, tracef);
+		if (nw == 1) {
 			tracef_size += nw;
 			fflush(tracef);
-		} else if (nw < 0) {
+		} else {
 			if (errno != EPIPE
 #if defined(EILSEQ) /*[*/
 					   && errno != EILSEQ
@@ -239,10 +304,12 @@ vwtrace(const char *fmt, va_list args)
 				stop_tracing();
 		}
 		if (tracef_pipe != NULL) {
-			nw = vfprintf(tracef_pipe, fmt, args);
-			if (nw < 0) {
+			nw = fwrite(buf, n2w, 1, tracef_pipe);
+			if (nw != 1) {
 				(void) fclose(tracef_pipe);
 				tracef_pipe = NULL;
+			} else {
+			    	fflush(tracef_pipe);
 			}
 		}
 	}
@@ -370,18 +437,23 @@ trace_rollover_check(void)
 #if defined(ROLLOVER_DEBUG) /*[*/
 		printf("final wpos = %lld\n", wpos);
 #endif /*]*/
+#if !defined(_MSC_VER) /*[*/
 		if (ftruncate(fileno(tracef), wpos) < 0) {
 			popup_an_errno(errno, "trace file ftruncate(%ld) "
 			    "failed", (long)wpos);
 			stop_tracing();
 			return;
 		}
+#endif /*]*/
 		if (fseeko(tracef, wpos, SEEK_SET) < 0) {
 			popup_an_errno(errno, "trace file fseeko(%ld) failed",
 			    (long)wpos);
 			stop_tracing();
 			return;
 		}
+#if defined(_MSC_VER) /*[*/
+		SetEndOfFile((HANDLE)_get_osfhandle(fileno(tracef)));
+#endif /*]*/
 		tracef_size = wpos;
 		tracef_midpoint = wpos;
 		Replace(tracef_midpoint_header,
@@ -409,19 +481,31 @@ create_tracefile_header(const char *mode)
 	clk = time((time_t *)0);
 	wtrace("Trace %s %s", mode, ctime(&clk));
 	wtrace(" Version: %s\n", build);
+	wtrace(" %s\n", build_options());
 	save_yourself();
 	wtrace(" Command: %s\n", command_string);
-	wtrace(" Model %s", model_name);
+	wtrace(" Model %s, %d rows x %d cols", model_name, maxROWS, maxCOLS);
 #if defined(X3270_DISPLAY) || (defined(C3270) && !defined(_WIN32)) /*[*/
 	wtrace(", %s display", appres.mono ? "monochrome" : "color");
 #endif /*]*/
 	if (appres.extended)
 		wtrace(", extended data stream");
-	wtrace(", %s emulation", appres.m3279 ? "color " : "monochrome");
+	wtrace(", %s emulation", appres.m3279 ? "color" : "monochrome");
 	wtrace(", %s charset", get_charset_name());
 	if (appres.apl_mode)
 		wtrace(", APL mode");
 	wtrace("\n");
+#if !defined(_WIN32) /*[*/
+	wtrace(" Locale codeset: %s\n", locale_codeset);
+#else /*][*/
+	wtrace(" ANSI codepage: %d\n", GetACP());
+#endif /*]*/
+	wtrace(" Host codepage: %d", (int)(cgcsgid & 0xffff));
+#if defined(X3270_DBCS) /*[*/
+	if (dbcs)
+		wtrace("+%d", (int)(cgcsgid_dbcs & 0xffff));
+#endif /*]*/
+	wtrace("\n");
 	if (CONNECTED)
 		wtrace(" Connected to %s, port %u\n",
 		    current_host, current_port);
@@ -556,7 +640,7 @@ get_devfd(const char *pathname)
 
 /* Callback for "OK" button on trace popup */
 static void
-tracefile_callback(Widget w, XtPointer client_data, XtPointer call_data unused)
+tracefile_callback(Widget w, XtPointer client_data, XtPointer call_data _is_unused)
 {
 	char *tfn = CN;
 	int devfd = -1;
@@ -657,6 +741,7 @@ tracefile_callback(Widget w, XtPointer client_data, XtPointer call_data unused)
 				Free(tfn);
 				return;
 			}
+			Replace(tracefile_name, NewString(tfn));
 			(void) SETLINEBUF(tracef);
 #if !defined(_WIN32) /*[*/
 			(void) fcntl(fileno(tracef), F_SETFD, 1);
@@ -691,11 +776,12 @@ tracefile_callback(Widget w, XtPointer client_data, XtPointer call_data unused)
 	}
 #endif /*]*/
 
-#if defined(_WIN32) /*[*/
+#if defined(_WIN32) && defined(C3270) /*[*/
 	/* Start the monitor window. */
 	if (tracef != stdout && appres.trace_monitor) {
 		STARTUPINFO startupinfo;
 		PROCESS_INFORMATION process_information;
+		char *path;
 		char *args;
 
 	    	(void) memset(&startupinfo, '\0', sizeof(STARTUPINFO));
@@ -703,10 +789,10 @@ tracefile_callback(Widget w, XtPointer client_data, XtPointer call_data unused)
 		startupinfo.lpTitle = tfn;
 		(void) memset(&process_information, '\0',
 			      sizeof(PROCESS_INFORMATION));
-		args = Malloc(32 + strlen(tfn));
-		sprintf(args, "catf.exe %s", tfn);
+		path = xs_buffer("%scatf.exe", instdir);
+		args = xs_buffer("\"%scatf.exe\" \"%s\"", instdir, tfn);
 		if (CreateProcess(
-		    "catf.exe",
+		    path,
 		    args,
 		    NULL,
 		    NULL,
@@ -716,10 +802,12 @@ tracefile_callback(Widget w, XtPointer client_data, XtPointer call_data unused)
 		    NULL,
 		    &startupinfo,
 		    &process_information) == 0) {
-		    	popup_an_error("CreateProcess(catf) failed: Windows "
-				"error %d", GetLastError());
+		    	popup_an_error("CreateProcess(%s) failed: %s",
+				path, win32_strerror(GetLastError()));
+			Free(path);
 			Free(args);
 		} else {
+			Free(path);
 		    	Free(args);
 			tracewindow_handle = process_information.hProcess;
 			CloseHandle(process_information.hThread);
@@ -750,7 +838,7 @@ tracefile_callback(Widget w, XtPointer client_data, XtPointer call_data unused)
 /* Callback for "No File" button on trace popup */
 static void
 no_tracefile_callback(Widget w, XtPointer client_data,
-	XtPointer call_data unused)
+	XtPointer call_data _is_unused)
 {
 	tracefile_callback((Widget)NULL, "", PN);
 	XtPopdown(trace_shell);
@@ -761,7 +849,7 @@ no_tracefile_callback(Widget w, XtPointer client_data,
 static void
 tracefile_on(int reason, enum toggle_type tt)
 {
-	char tracefile_buf[256];
+	char *tracefile_buf = NULL;
 	char *tracefile;
 
 	if (tracef != (FILE *)NULL)
@@ -776,9 +864,10 @@ tracefile_on(int reason, enum toggle_type tt)
 		tracefile = appres.trace_file;
 	else {
 #if defined(_WIN32) /*[*/
-		(void) sprintf(tracefile_buf, "x3trc.%d.txt", getpid());
+		tracefile_buf = xs_buffer("%sx3trc.%u.txt", myappdata,
+			getpid());
 #else /*][*/
-		(void) sprintf(tracefile_buf, "%s/x3trc.%d", appres.trace_dir,
+		tracefile_buf = xs_buffer("%s/x3trc.%u", appres.trace_dir,
 			getpid());
 #endif /*]*/
 		tracefile = tracefile_buf;
@@ -789,6 +878,8 @@ tracefile_on(int reason, enum toggle_type tt)
 #endif /*]*/
 	{
 		tracefile_callback((Widget)NULL, tracefile, PN);
+		if (tracefile_buf != NULL)
+		    	Free(tracefile_buf);
 		return;
 	}
 #if defined(X3270_DISPLAY) /*[*/
@@ -808,6 +899,9 @@ tracefile_on(int reason, enum toggle_type tt)
 
 	popup_popup(trace_shell, XtGrabExclusive);
 #endif /*]*/
+
+	if (tracefile_buf != NULL)
+		Free(tracefile_buf);
 }
 
 /* Close the trace file. */
@@ -834,7 +928,7 @@ tracefile_off(void)
 }
 
 void
-toggle_dsTrace(struct toggle *t unused, enum toggle_type tt)
+toggle_dsTrace(struct toggle *t _is_unused, enum toggle_type tt)
 {
 	/* If turning on trace and no trace file, open one. */
 	if (toggled(DS_TRACE) && tracef == NULL)
@@ -850,7 +944,7 @@ toggle_dsTrace(struct toggle *t unused, enum toggle_type tt)
 }
 
 void
-toggle_eventTrace(struct toggle *t unused, enum toggle_type tt)
+toggle_eventTrace(struct toggle *t _is_unused, enum toggle_type tt)
 {
 	/* If turning on event debug, and no trace file, open one. */
 	if (toggled(EVENT_TRACE) && tracef == NULL)
@@ -877,7 +971,7 @@ do_screentrace(void)
 {
 	register int i;
 
-	if (fprint_screen(screentracef, False, False)) {
+	if (fprint_screen(screentracef, False, P_TEXT)) {
 		for (i = 0; i < COLS; i++)
 			(void) fputc('=', screentracef);
 		(void) fputc('\n', screentracef);
@@ -952,8 +1046,8 @@ screentrace_cb(char *tfn)
 #if defined(X3270_DISPLAY) /*[*/
 /* Callback for "OK" button on screentrace popup */
 static void
-screentrace_callback(Widget w unused, XtPointer client_data,
-    XtPointer call_data unused)
+screentrace_callback(Widget w _is_unused, XtPointer client_data,
+    XtPointer call_data _is_unused)
 {
 	if (screentrace_cb(XawDialogGetValueString((Widget)client_data)))
 		XtPopdown(screentrace_shell);
@@ -961,7 +1055,7 @@ screentrace_callback(Widget w unused, XtPointer client_data,
 
 /* Callback for second "OK" button on screentrace popup */
 static void
-onescreen_callback(Widget w, XtPointer client_data, XtPointer call_data unused)
+onescreen_callback(Widget w, XtPointer client_data, XtPointer call_data _is_unused)
 {
 	char *tfn;
 
@@ -992,9 +1086,9 @@ onescreen_callback(Widget w, XtPointer client_data, XtPointer call_data unused)
 #endif /*]*/
 
 void
-toggle_screenTrace(struct toggle *t unused, enum toggle_type tt)
+toggle_screenTrace(struct toggle *t _is_unused, enum toggle_type tt)
 {
-	char tracefile_buf[256];
+	char *tracefile_buf = NULL;
 	char *tracefile;
 
 	if (toggled(SCREEN_TRACE)) {
@@ -1002,16 +1096,18 @@ toggle_screenTrace(struct toggle *t unused, enum toggle_type tt)
 			tracefile = appres.screentrace_file;
 		else {
 #if defined(_WIN32) /*[*/
-			(void) sprintf(tracefile_buf, "x3scr.%d.txt",
-				       getpid());
+			tracefile_buf = xs_buffer("%sx3scr.%u.txt",
+				myappdata, getpid());
 #else /*][*/
-			(void) sprintf(tracefile_buf, "%s/x3scr.%d",
+			tracefile_buf = xs_buffer("%s/x3scr.%u",
 				appres.trace_dir, getpid());
 #endif /*]*/
 			tracefile = tracefile_buf;
 		}
 		if (tt == TT_INITIAL || tt == TT_ACTION) {
 			(void) screentrace_cb(NewString(tracefile));
+			if (tracefile_buf != NULL)
+				Free(tracefile_buf);
 			return;
 		}
 #if defined(X3270_DISPLAY) /*[*/
@@ -1033,6 +1129,9 @@ toggle_screenTrace(struct toggle *t unused, enum toggle_type tt)
 			do_screentrace();
 		(void) fclose(screentracef);
 	}
+
+	if (tracefile_buf != NULL)
+		Free(tracefile_buf);
 }
 
 #endif /*]*/
